import os

from raichu.substrate import PksStarterSubstrate
from raichu.cluster.modular_cluster import ModularCluster
from raichu.cluster.ripp_cluster import RiPPCluster
from raichu.cluster.terpene_cluster import TerpeneCluster
from raichu.domain.domain import (
    TailoringDomain,
    CarrierDomain,
    SynthesisDomain,
    RecognitionDomain,
    TerminationDomain,
    UnknownDomain,
    Domain,
)
from raichu.module import (
    PKSModuleSubtype,
    NRPSModule,
    LinearPKSModule,
    IterativePKSModule,
    TransATPKSModule,
    ModuleType,
)
from raichu.domain.domain_types import (
    TailoringDomainType,
    TerminationDomainType,
    CarrierDomainType,
    SynthesisDomainType,
    RecognitionDomainType,
)
from raichu.cluster.alkaloid_cluster import AlkaloidCluster
from raichu.tailoring_enzymes import TailoringEnzyme, TailoringEnzymeType
from raichu.representations import (
    ClusterRepresentation,
    DomainRepresentation,
    ModuleRepresentation,
    TailoringRepresentation,
    CleavageSiteRepresentation,
    MacrocyclizationRepresentation,
)

DOMAIN_TO_SUPERTYPE = {}
for domain_name in TailoringDomainType.__members__:
    DOMAIN_TO_SUPERTYPE[domain_name] = TailoringDomain
for domain_name in CarrierDomainType.__members__:
    DOMAIN_TO_SUPERTYPE[domain_name] = CarrierDomain
for domain_name in SynthesisDomainType.__members__:
    DOMAIN_TO_SUPERTYPE[domain_name] = SynthesisDomain
for domain_name in RecognitionDomainType.__members__:
    DOMAIN_TO_SUPERTYPE[domain_name] = RecognitionDomain
for domain_name in TerminationDomainType.__members__:
    DOMAIN_TO_SUPERTYPE[domain_name] = TerminationDomain
DOMAIN_TO_SUPERTYPE["UNKNOWN"] = UnknownDomain


def make_domain(
    domain_repr: DomainRepresentation, substrate: str, strict: bool = True
) -> Domain:
    domain_class = DOMAIN_TO_SUPERTYPE.get(domain_repr.type)

    if not domain_repr.name:
        domain_repr.name = domain_repr.type
    if domain_class:
        if domain_class == RecognitionDomain:
            domain = domain_class(
                domain_repr.type,
                substrate,
                domain_subtype=domain_repr.subtype,
                active=domain_repr.active,
                used=domain_repr.used,
            )
        elif domain_class == UnknownDomain:
            domain = UnknownDomain(domain_repr.name)
        else:
            domain = domain_class(
                domain_repr.type,
                domain_subtype=domain_repr.subtype,
                active=domain_repr.active,
                used=domain_repr.used,
            )
    elif strict:
        raise ValueError(f"Unrecognised domain type: {domain_repr.type}")
    else:
        domain = UnknownDomain(domain_repr.name)

    return domain


def build_cluster(
    cluster_repr: ClusterRepresentation, strict: bool = True
) -> ModularCluster:

    genes = set()
    new_starter = False
    modules = []
    previous_domain = None

    if len(cluster_repr.modules) == 0:
        raise ValueError(
            "Cluster is empty.This can happen with Type III PKS clusters. Please check the input file."
        )

    for i, module_repr in enumerate(cluster_repr.modules):
        if i == 0 or new_starter:
            starter = True
            if new_starter:

                if not strict and module_repr.type == 'PKS' and module_repr.subtype == 'PKS_CIS':
                    if "AT" in [domain.type for domain in module_repr.domains] and module_repr.substrate not in [v.name for v in PksStarterSubstrate]:
                        module_repr.substrate = "WILDCARD"

            new_starter = False

        else:
            starter = False

        if i == len(cluster_repr.modules) - 1:
            terminator = True
        else:
            terminator = False

        domains = []
        for domain_repr in module_repr.domains:
            domain = make_domain(domain_repr, module_repr.substrate, strict=strict)

            if domain_repr.gene_name is not None:
                if previous_domain:
                    previous_gene = previous_domain.gene
                    if (
                        previous_gene != domain_repr.gene_name
                        and domain_repr.gene_name in genes
                    ):
                        raise ValueError(
                            f"Gene name '{previous_gene}' already assigned to upstream domain(s)."
                        )

                domain.set_gene(domain_repr.gene_name)
                genes.add(domain_repr.gene_name)
            domains.append(domain)
            previous_domain = domain
        module_type = ModuleType.from_string(module_repr.type)
        if module_type.name == "PKS":
            if module_repr.subtype is not None:
                module_subtype = PKSModuleSubtype.from_string(module_repr.subtype)

                if module_subtype.name == "PKS_CIS":
                    module = LinearPKSModule(
                        i, domains, starter=starter, terminator=terminator
                    )
                elif module_subtype.name == "PKS_TRANS":
                    module = TransATPKSModule(
                        i, domains, starter=starter, terminator=terminator
                    )
                elif module_subtype.name == "PKS_ITER":
                    module = IterativePKSModule(
                        i,
                        domains,
                        starter=starter,
                        terminator=terminator,
                        iterations=module_repr.iterations,
                    )
                else:
                    raise ValueError(
                        f"Unrecognised PKS module subtype: {module_subtype}."
                    )
            else:
                raise ValueError("PKS module subtype must be specified.")

        elif module_type.name == "NRPS":
            if module_repr.subtype is None:
                module = NRPSModule(i, domains, starter=starter, terminator=terminator)
            else:
                raise ValueError(
                    "NRPS module subtypes are currently not supported. Please pass None."
                )
        else:
            raise ValueError(f"Unrecognised module type: {module_repr.type}")

        if (
            not any(
                [domain.type.name in ["CP", "ACP", "PCP"] for domain in module.domains]
            )
            and strict
        ):
            raise ValueError(
                f"Module {module.id} does not contain a carrier protein domain."
            )

        if module.is_broken and module.is_starter_module:
            new_starter = True

        modules.append(module)
    cluster = ModularCluster(modules, cluster_repr.tailoring_enzymes)
    has_functional_modules = False
    for module in cluster.modules:
        if not module.is_broken:
            has_functional_modules = True
    if not has_functional_modules:
        print("WARNING: Cluster has no functional modules. Drawing this cluster will result in errors.")

    return cluster


def draw_cluster(
    cluster_repr: ClusterRepresentation, out_file=None, colour_by_module=True
) -> None:
    cluster = build_cluster(cluster_repr, strict=False)
    cluster.compute_structures(compute_cyclic_products=False)
    cluster.do_tailoring()

    if out_file:
        return cluster.draw_cluster(
            as_string=False, out_file=out_file, colour_by_module=colour_by_module
        )
    else:
        return cluster.draw_cluster()


def draw_products(cluster_repr: ClusterRepresentation, out_dir) -> None:
    cluster = build_cluster(cluster_repr)
    cluster.compute_structures(compute_cyclic_products=False)
    cluster.do_tailoring()
    cluster.cyclise_all()
    cluster.draw_all_products(out_dir)


def draw_product(cluster_repr: ClusterRepresentation, out_file) -> None:
    cluster = build_cluster(cluster_repr)
    cluster.compute_structures(compute_cyclic_products=True)
    cluster.do_tailoring()
    cluster.draw_product(as_string=False, out_file=out_file)


def draw_ripp_structure(ripp_cluster: RiPPCluster, out_folder: str) -> None:
    if not os.path.exists(out_folder):
        os.mkdir(out_folder)

    ripp_cluster.make_peptide()
    ripp_cluster.draw_product(
        as_string=False, out_file=os.path.join(out_folder, "peptide_test_ripp.svg")
    )
    ripp_cluster.draw_cluster(
        as_string=False, out_file=os.path.join(out_folder, "ripp_inline.svg")
    )

    order = []

    ripp_cluster.do_tailoring()
    if ripp_cluster.tailoring_representations:
        ripp_cluster.draw_tailoring(
            out_file=os.path.join(out_folder, "ripp_tailoring_steps.svg")
        )
        ripp_cluster.draw_product(
            as_string=False, out_file=os.path.join(out_folder, "ripp_tailoring.svg")
        )
        order.append("tailoring")
    if ripp_cluster.macrocyclisation_representations:
        ripp_cluster.do_macrocyclization()
        ripp_cluster.draw_product(
            as_string=False,
            out_file=os.path.join(out_folder, "ripp_macrocyclisation.svg"),
        )
        order.append("cyclisation")

    if ripp_cluster.cleavage_sites:
        ripp_cluster.do_proteolytic_cleavage()
        ripp_cluster.draw_product(
            as_string=False, out_file=os.path.join(out_folder, "cleavage_test_ripp.svg")
        )
        order.append("cleavage")

    ripp_cluster.draw_pathway(
        out_file=os.path.join(out_folder, "ripp_pathway.svg"), order=tuple(order)
    )


def compute_all_ripp_masses(
    precursor: str, out_file: str, min_length=3, max_length=100
) -> None:
    def find_substrings(input_string) -> list[str]:
        # Initialize an empty set to store unique substrings
        substrings = set()

        # Get the length of the input string
        n = len(input_string)

        # Outer loop for substring start index
        for i in range(n):
            # Inner loop for substring end index
            for j in range(i + min_length, min(i + max_length + 1, n + 1)):
                # Slice the string from i to j
                substring = input_string[i:j]

                # Add the substring to the set
                substrings.add(substring)

        # Convert the set to a list to return
        return list(substrings)

    entries = []

    core_peptides = find_substrings(precursor)
    for core_peptide in core_peptides:
        ripp_cluster = RiPPCluster("placeholder_gene", precursor, core_peptide)
        ripp_cluster.make_peptide()
        mass = ripp_cluster.linear_product.get_mass()
        formatted_precursor = precursor.lower().replace(
            core_peptide.lower(), core_peptide
        )

        entries.append([core_peptide, formatted_precursor, mass, mass + 1.007276])
    # Sort based on mass
    sorted_entries = sorted(entries, key=lambda x: x[2])
    with open(out_file, "w", newline="") as csvfile:
        csvwriter = csv.writer(csvfile)
        csvwriter.writerow(
            ["CorePeptide", "PrecursorWithCorePeptide", "Mass", "[M+H]+"]
        )

        for entry in sorted_entries:
            csvwriter.writerow([entry[0], entry[1], str(entry[2]), str(entry[3])])


def draw_terpene_structure(terpene_cluster: TerpeneCluster) -> None:
    terpene_cluster.create_precursor()
    terpene_cluster.draw_product(as_string=False, out_file="precursor_test_terpene.svg")
    terpene_cluster.do_macrocyclization()
    terpene_cluster.draw_product(
        as_string=False, out_file="macroyclisation_test_terpene.svg"
    )
    terpene_cluster.do_tailoring()
    terpene_cluster.draw_product(as_string=False, out_file="tailoring_test_terpene.svg")
    print(get_tailoring_sites(terpene_cluster.chain_intermediate))


def draw_alkaloid_structure(alkaloid_cluster: AlkaloidCluster) -> None:
    alkaloid_cluster.make_scaffold()
    alkaloid_cluster.draw_product(
        as_string=False, out_file="precursor_test_alkaloid.svg"
    )
    alkaloid_cluster.do_tailoring()
    alkaloid_cluster.draw_product(
        as_string=False, out_file="tailoring_test_alkaloid.svg"
    )


def get_spaghettis(cluster_repr: ClusterRepresentation) -> list[str]:

    cluster = build_cluster(cluster_repr)
    cluster.compute_structures(compute_cyclic_products=False)
    cluster.do_tailoring()
    cluster.draw_cluster()
    spaghettis = cluster.get_spaghettis()

    return spaghettis


def get_tailoring_sites(structure, enzyme_name="all", out_file=None):
    tailoring_sites = {}
    if enzyme_name == "all":
        for enzyme_type in TailoringEnzymeType:
            tailoring_enzyme = TailoringEnzyme("gene", enzyme_type.name)
            tailoring_sites[enzyme_type.name] = tailoring_enzyme.get_possible_sites(
                structure
            )
    else:
        for enzyme_type in TailoringEnzymeType:
            if enzyme_type.name == enzyme_name:
                tailoring_enzyme = TailoringEnzyme("gene", enzyme_type.name)
                tailoring_sites[enzyme_type.name] = tailoring_enzyme.get_possible_sites(
                    structure, out_file=out_file
                )

    return tailoring_sites


def get_tailoring_sites_atom_names(structure):
    tailoring_sites = {}
    for enzyme_type in TailoringEnzymeType:
        tailoring_enzyme = TailoringEnzyme("gene", enzyme_type.name)
        tailoring_sites[enzyme_type.name] = [
            str(atom) if type(atom) != list else [str(subatom) for subatom in atom]
            for atom in tailoring_enzyme.get_possible_sites(structure)
        ]
    return tailoring_sites


def read_in_cluster(input_dir):
    cluster_representation = ClusterRepresentation.from_file(input_dir)
    print(cluster_representation)
    cluster = build_cluster(cluster_representation, strict=False)
    return cluster, cluster_representation


if __name__ == "__main__":
    # atropopeptide

    ripp_cluster = RiPPCluster(
        "best_ripp(tryptorubin)_encoding_gene",
        "mkaekslkayawyiwyaha",
        "slkayawyiwy",
        cleavage_sites=[CleavageSiteRepresentation("Y", 10, "end")],
        tailoring_representations=[
            TailoringRepresentation(
                "p450", "DOUBLE_BOND_REDUCTASE", [["C_139", "C_138"]]
            ),
            TailoringRepresentation(
                "p450",
                "OXIDATIVE_BOND_SYNTHASE",
                [["C_139", "N_134"], ["C_120", "N_102"], ["C_138", "C_107"]],
            ),
        ],
    )

    lanthipeptide_type_I_cluster_catenulipeptin = RiPPCluster(
        "Caci_4240",
        "MTEEMTLLDLQGMEQTETDSWGGSGHGGGGDSGLSVTGCNGHSGISLLCDL",
        "GHGGGGDSGLSVTGCNGHSGISLLCDL",
        tailoring_representations=[
            TailoringRepresentation(
                "Caci_4239", "THREONINE_SERINE_DEHYDRATASE", [["O_64"], ["O_144"]]
            ),
            TailoringRepresentation(
                "Caci_4239",
                "LANTHIPEPTIDE_CYCLASE",
                [
                    ["C_43", "C_63"],
                    ["S_92", "C_63"],
                    ["C_122", "C_143"],
                    ["S_169", "C_143"],
                ],
            ),
        ],
    )
    lanthipeptide_type_III_cluster_catenulipeptin = RiPPCluster(
        "Caci_4240",
        "MTEEMTLLDLQGMEQTETDSWGGSGHGGGGDSGLSVTGCNGHSGISLLCDL",
        "GHGGGGDSGLSVTGCNGHSGISLLCDL",
        tailoring_representations=[
            TailoringRepresentation(
                "Caci_4239",
                "LANTHIONINE_SYNTHETASE",
                [
                    ["C_43", "C_63"],
                    ["S_92", "C_63"],
                    ["C_122", "C_143"],
                    ["S_169", "C_143"],
                ],
            )
        ],
    )
    proteusins_cluster_polytheonamide_a = RiPPCluster(
        "poyA",
        "MADSDNTPTSRKDFETAIIAKAWKDPEYLRRLRSNPREVLQEELEALHPGAQLPDDLGISIHEEDENHVHLVMPRHPQNVSDQTLTDDDLDQAAGGTGIGVVVAVVAGAVANTGAGVNQVAGGNINVVGNINVNANVSVNMNQTT",
        "TGIGVVVAVVAGAVANTGAGVNQVAGGNINVVGNINVNANVSVNMNQTT",
        tailoring_representations=[
            TailoringRepresentation(
                "rSAM epimerase",
                "AMINO_ACID_EPIMERASE",
                [
                    ["C_269"],
                    ["C_144"],
                    ["C_284"],
                    ["C_163"],
                    ["C_36"],
                    ["C_301"],
                    ["C_52"],
                    ["C_90"],
                    ["C_185"],
                    ["C_316"],
                    ["C_66"],
                    ["C_204"],
                    ["C_334"],
                    ["C_84"],
                    ["C_221"],
                    ["C_353"],
                    ["C_104"],
                    ["C_233"],
                    ["C_252"],
                ],
            )
        ],
    )
    sliceotide_cluster = RiPPCluster(
        "plpA",
        "NVSVNMNQTTRVSVNMNQTNVSVNVSVNMNQTNVSVNVSVNMNQTNVSVNVSVN",
        "NVSVNMNQTTR",
        tailoring_representations=[
            TailoringRepresentation("rSAM epimerase", "SPLICEASE", [["C_30", "C_22"]]),
            TailoringRepresentation("arginase", "ARGINASE", [["N_93"]]),
        ],
    )
    thiopeptide_cluster_thiomuracin = RiPPCluster(
        "tpdA",
        "MDLSDLPMDVFELADDGVAVESLTAGHGMTEVGASCNCFCYICCSCSSA",
        "SCNCFCYICCSCSS",
        tailoring_representations=[
            TailoringRepresentation(
                "tpdD",
                "THREONINE_SERINE_DEHYDRATASE",
                [["O_104"], ["O_90"], ["O_4"], ["O_111"]],
            ),
            TailoringRepresentation(
                "tpdE",
                "CYCLODEHYDRASE",
                [["S_97"], ["S_11"], ["S_76"], ["S_46"], ["S_83"], ["S_27"]],
            ),
            TailoringRepresentation("tpdF", "THIOPEPTIDE_CYCLASE", [["C_3", "C_89"]]),
            TailoringRepresentation("tpdF", "METHYLTRANSFERASE", [["C_26"]]),
            TailoringRepresentation("tpdF", "HYDROXYLASE", [["C_33"]]),
            TailoringRepresentation(
                "tpdF", "DEHYDROGENASE", [["C_67", "C_68"], ["C_3", "C_89"]]
            ),
            TailoringRepresentation("tpdF", "EPOXIDASE", [["C_67", "C_68"]]),
            TailoringRepresentation("tpdF", "MONOAMINE_OXIDASE", [["N_0"]]),
            TailoringRepresentation("tpdF", "DEHYDRATASE", [["C_1", "C_84"]]),
        ],
    )
    cyanobactin_cluster_trunkamide = RiPPCluster(
        "truE",
        "MNKKNILPQLGQPVIRLTAGQLSSQLAELSEEALGGVDASTSIAPFCSYDGVDASSYDGVDASSYDD",
        "TSIAPFC",
        macrocyclisations=[MacrocyclizationRepresentation("N_0", "O_59")],
        tailoring_representations=[
            TailoringRepresentation("truD", "CYCLODEHYDRASE", [["S_56"]]),
            TailoringRepresentation(
                "truF", "PRENYLTRANSFERASE", [["O_13"], ["O_5"]], "3_METHYL_1_BUTENYL"
            ),
        ],
    )

    lasso_peptide_cluster = RiPPCluster(
        "A1S42_RS12075",
        "MKYCKPTFESIATFKKDTKGLWTGKFRDIFGGRAIVRIRIEF",
        "MKYCKPTFESIATFKKDTKGLWTGKFRDIFGGRAIVRIRIEF",
        tailoring_representations=[
            TailoringRepresentation("lasB", "PROTEASE", [["N_180", "C_178"]]),
            TailoringRepresentation("lasC", "MACROLACTAM_SYNTHETASE", [["O_261"]]),
        ],
    )
    sacti_peptide_cluster_thurincin = RiPPCluster(
        "thnA",
        "METPVVQPRDWTCWSCLVCAACSVELLNLVTAATGASTAS",
        "DWTCWSCLVCAACSVELLNLVTAATGASTAS",
        tailoring_representations=[
            TailoringRepresentation(
                "thnB",
                "OXIDATIVE_BOND_SYNTHASE",
                [
                    ["S_109", "C_222"],
                    ["S_66", "C_203"],
                    ["S_90", "C_182"],
                    ["S_37", "C_156"],
                ],
            )
        ],
    )

    terpene_cluster = TerpeneCluster(
        "limonene_synthase",
        "GERANYL_PYROPHOSPHATE",
        macrocyclisations=[MacrocyclizationRepresentation("C_13", "C_8")],
        cyclase_type="Class_1",
        tailoring_representations=[
            TailoringRepresentation(
                "pseudo_isomerase",
                "DOUBLE_BOND_SHIFT",
                [["C_13", "C_14", "C_14", "C_15"]],
            )
        ],
    )

    alkaloid_cluster = AlkaloidCluster(
        "phenylalanine",
        tailoring_representations=[
            TailoringRepresentation("pseudo_decarboxylase", "DECARBOXYLASE", [["C_9"]]),
            TailoringRepresentation(
                "pseudo_hydroxylase", "PRENYLTRANSFERASE", [["C_7"]], "DIMETHYLALLYL"
            ),
            TailoringRepresentation(
                "pseudo_decarboxylase", "HALOGENASE", [["C_10"]], "Cl"
            ),
            TailoringRepresentation("pseudo_hydroxylase", "HYDROXYLASE", [["C_6"]]),
            TailoringRepresentation(
                "methyltransferase", "METHYLTRANSFERASE", [["N_12"], ["C_7"], ["O_25"]]
            ),
        ],
    )

    cluster_repr = ClusterRepresentation(
        [
            ModuleRepresentation(
                "PKS",
                "PKS_CIS",
                "ACETYL_COA",
                [
                    DomainRepresentation("Gene 1", "AT", None, None, True, True),
                    DomainRepresentation("Gene 1", "ACP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "PKS",
                "PKS_ITER",
                "METHYLMALONYL_COA",
                [
                    DomainRepresentation("Gene 1", "KS", None, None, True, True),
                    DomainRepresentation("Gene 1", "AT", None, None, True, True),
                    DomainRepresentation("Gene 1", "AT", None, None, True, False),
                    DomainRepresentation("Gene 1", "DH", None, None, True, True),
                    DomainRepresentation("Gene 1", "ER", None, None, True, True),
                    DomainRepresentation("Gene 1", "ACP", None, None, True, True),
                ],
                5,
            ),
            ModuleRepresentation(
                "PKS",
                "PKS_CIS",
                "METHYLMALONYL_COA",
                [
                    DomainRepresentation("Gene 1", "KS", None, None, True, True),
                    DomainRepresentation("Gene 1", "AT", None, None, True, True),
                    DomainRepresentation("Gene 1", "TE", None, None, True, True),
                ],
            ),
        ]
    )

    trans_at_ks_cluster_repr = ClusterRepresentation(
        [
            ModuleRepresentation(
                "PKS",
                "PKS_TRANS",
                "ACETYL_COA",
                [
                    DomainRepresentation("Gene 1", "AT", None, None, True, True),
                    DomainRepresentation("Gene 1", "ACP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "PKS",
                "PKS_TRANS",
                "METHYLMALONYL_COA",
                [
                    DomainRepresentation("Gene 1", "KS", "BETA_D_OH", None, True, True),
                    DomainRepresentation("Gene 1", "DH", None, None, True, True),
                    DomainRepresentation("Gene 1", "ER", None, None, True, True),
                    DomainRepresentation("Gene 1", "ACP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "PKS",
                "PKS_TRANS",
                "METHYLMALONYL_COA",
                [
                    DomainRepresentation("Gene 1", "KS", "BETA_D_OH", None, True, True),
                    DomainRepresentation("Gene 1", "DH", None, None, True, True),
                    DomainRepresentation("Gene 1", "ER", None, None, True, True),
                    DomainRepresentation("Gene 1", "ACP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "PKS",
                "PKS_TRANS",
                "METHYLMALONYL_COA",
                [
                    DomainRepresentation("Gene 1", "KS", "BETA_D_OH", None, True, True),
                    DomainRepresentation("Gene 1", "DH", None, None, True, True),
                    DomainRepresentation("Gene 1", "ER", None, None, True, True),
                    DomainRepresentation("Gene 1", "ACP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "PKS",
                "PKS_TRANS",
                "METHYLMALONYL_COA",
                [
                    DomainRepresentation("Gene 1", "KS", "PYR", None, True, True),
                    DomainRepresentation("Gene 1", "AT", None, None, True, True),
                    DomainRepresentation("Gene 1", "AT", None, None, True, False),
                    DomainRepresentation("Gene 1", "DH", None, None, True, True),
                    DomainRepresentation("Gene 1", "ER", None, None, True, True),
                    DomainRepresentation("Gene 1", "ACP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "PKS",
                "PKS_CIS",
                "METHYLMALONYL_COA",
                [
                    DomainRepresentation("Gene 1", "KS", None, None, True, True),
                    DomainRepresentation("Gene 1", "AT", None, None, True, True),
                    DomainRepresentation("Gene 1", "TE", None, None, True, True),
                ],
            ),
        ]
    )
    # draw_cluster(trans_at_ks_cluster_repr, outfile = "iterative_pks.svg")
    # draw_ripp_structure(ripp_cluster)

    cluster_nrps = ClusterRepresentation(
        [
            ModuleRepresentation(
                "NRPS",
                None,
                "valine",
                [
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "NRPS",
                None,
                "proline",
                [
                    DomainRepresentation("Gene 1", "C", None, None, True, True),
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "NRPS",
                None,
                "threonine",
                [
                    DomainRepresentation("Gene 1", "C", None, None, True, True),
                    DomainRepresentation("Gene 1", "CYC", None, None, True, True),
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "OX", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "NRPS",
                None,
                "proline",
                [
                    DomainRepresentation("Gene 1", "C", None, None, True, True),
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                    DomainRepresentation("Gene 1", "TE", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "NRPS",
                None,
                "valine",
                [
                    DomainRepresentation("Gene 1", "C", None, None, True, True),
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "NRPS",
                None,
                "valine",
                [
                    DomainRepresentation("Gene 1", "C", None, None, True, True),
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                ],
            ),
        ]
    )

    nrps_cluster = ClusterRepresentation(
        [
            ModuleRepresentation(
                "NRPS",
                None,
                "serine",
                [
                    DomainRepresentation("Gene 1", "C", None, None, True, True),
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "NRPS",
                None,
                "threonine",
                [
                    DomainRepresentation("Gene 1", "C", None, None, True, True),
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "NRPS",
                None,
                "tryptophan",
                [
                    DomainRepresentation("Gene 1", "C", None, None, True, True),
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "NRPS",
                None,
                "aspartic acid",
                [
                    DomainRepresentation("Gene 1", "C", None, None, True, True),
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "NRPS",
                None,
                "aspartic acid",
                [
                    DomainRepresentation("Gene 1", "C", None, None, True, True),
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "NRPS",
                None,
                "4-hydroxyphenylglycine",
                [
                    DomainRepresentation("Gene 1", "C", None, None, True, True),
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "NRPS",
                None,
                "aspartic acid",
                [
                    DomainRepresentation("Gene 1", "C", None, None, True, True),
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "NRPS",
                None,
                "alanine",
                [
                    DomainRepresentation("Gene 1", "C", None, None, True, True),
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "NRPS",
                None,
                "asparagine",
                [
                    DomainRepresentation("Gene 1", "C", None, None, True, True),
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "NRPS",
                None,
                "glutamic acid",
                [
                    DomainRepresentation("Gene 1", "C", None, None, True, True),
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                ],
            ),
            ModuleRepresentation(
                "NRPS",
                None,
                "tryptophan",
                [
                    DomainRepresentation("Gene 1", "C", None, None, True, True),
                    DomainRepresentation("Gene 1", "A", None, None, True, True),
                    DomainRepresentation("Gene 1", "PCP", None, None, True, True),
                ],
            ),
        ]
    )
    draw_cluster(nrps_cluster, out_file="antismash_test.svg")
    # draw_ripp_structure(ripp_cluster)
    # ripp_cluster.draw_precursor(as_string= False, out_file= "bubbles.svg")

    # cyanobactin_cluster_trunkamide.make_peptide()
    # # print(get_tailoring_sites_atom_names(
    # #     cyanobactin_cluster_trunkamide.chain_intermediate))
    # cyanobactin_cluster_trunkamide.draw_product(
    #     as_string=False, out_file="peptide_test_cyanobactin_peptide.svg")
    # cyanobactin_cluster_trunkamide.do_tailoring()
    # cyanobactin_cluster_trunkamide.draw_product(
    #     as_string=False, out_file="tailored_test_cyanobactin_peptide.svg", draw_straightened=False)
    # cyanobactin_cluster_trunkamide.do_macrocyclization()
    # cyanobactin_cluster_trunkamide.draw_product(
    #     as_string=False, out_file="final_peptide_test_cyanobactin_peptide.svg", draw_straightened=False)
    # lanthipeptide_type_I_cluster_catenulipeptin.make_peptide()
    # # print(get_tailoring_sites_atom_names(
    # #     lanthipeptide_type_III_cluster_catenulipeptin.chain_intermediate))
    # lanthipeptide_type_I_cluster_catenulipeptin.draw_product(
    #     as_string=False, out_file="peptide_test_lanthipeptide_peptide.svg")
    # lanthipeptide_type_I_cluster_catenulipeptin.do_tailoring()
    # lanthipeptide_type_I_cluster_catenulipeptin.draw_product(
    #     as_string=False, out_file="tailored_test_lanthipeptide_peptide.svg", draw_straightened=False)

    # thiopeptide_cluster_thiomuracin.make_peptide()
    #
    # thiopeptide_cluster_thiomuracin.draw_product(
    #     as_string=False, out_file="peptide_test_thiopeptide_peptide.svg")
    # thiopeptide_cluster_thiomuracin.do_tailoring()
    # thiopeptide_cluster_thiomuracin.draw_product(
    #     as_string=False, out_file="tailored_test_thiopeptide_peptide.svg")
    # get_tailoring_sites(thiopeptide_cluster_thiomuracin.chain_intermediate,
    #                     enzyme_name="DEHYDROGENASE", out_file="methylation_options_thiopeptide.svg")
    # # print(get_tailoring_sites_atom_names(
    # #     thiopeptide_cluster_thiomuracin.chain_intermediate))

    # get_tailoring_sites(thiopeptide_cluster_thiomuracin.chain_intermediate, enzyme_name="CYCLODEHYDRASE",
    #                     out_file="cyclodehydration.svg")
    # thiopeptide_cluster_thiomuracin.draw_precursor_with_modified_product(as_string=False, out_file="bubbles.svg")

    # print(get_tailoring_sites_atom_names(
    #     thiopeptide_cluster_thiomuracin.chain_intermediate))
    # sacti_peptide_cluster_thurincin.make_peptide()
    # # # print(get_tailoring_sites_atom_names(
    # # # sancti_peptide_cluster_thurincin.chain_intermediate))
    # # # sancti_peptide_cluster_thurincin.draw_product(
    # # #     as_string=False, out_file="peptide_test_sancti_peptide.svg")
    # sacti_peptide_cluster_thurincin.do_tailoring()
    # drawer = Drawer(sacti_peptide_cluster_thurincin.chain_intermediate)
    # drawer.show_molecule()

    # # sancti_peptide_cluster_thurincin.draw_product(
    # #     as_string=False, out_file="tailored_test_sancti_peptide.svg", draw_straightened=False)
    # # print(get_tailoring_sites_atom_names(
    # #     sancti_peptide_cluster_thurincin.chain_intermediate))
    # # proteusins_cluster_polytheonamide_a.make_peptide()
    # # proteusins_cluster_polytheonamide_a.draw_product(
    # #     as_string=False, out_file="peptide_test_proteusin_peptide.svg")
    # # proteusins_cluster_polytheonamide_a.do_tailoring()
    # # proteusins_cluster_polytheonamide_a.draw_product(
    # #     as_string=False, out_file="tailored_test_proteusin_peptide.svg", draw_straightened=False)
    # print(get_tailoring_sites_atom_names(
    #     thiopeptide_cluster_thiomuracin.chain_intermediate))

    # sancti_peptide_cluster_thurincin.make_peptide()
    # sancti_peptide_cluster_thurincin.draw_precursor_with_modified_product(as_string=False, out_file="bubbles.svg")
    # print(get_tailoring_sites_atom_names(
    # sancti_peptide_cluster_thurincin.chain_intermediate))
    # sancti_peptide_cluster_thurincin.draw_product(
    #     as_string=False, out_file="peptide_test_sancti_peptide.svg")
    # sancti_peptide_cluster_thurincin.do_tailoring()
    # sancti_peptide_cluster_thurincin.draw_product(
    #     as_string=False, out_file="tailored_test_sancti_peptide.svg", draw_straightened=False)
    # print(get_tailoring_sites_atom_names(
    #     sancti_peptide_cluster_thurincin.chain_intermediate))
    # proteusins_cluster_polytheonamide_a.make_peptide()
    # proteusins_cluster_polytheonamide_a.draw_product(
    #     as_string=False, out_file="peptide_test_proteusin_peptide.svg")
    # proteusins_cluster_polytheonamide_a.do_tailoring()
    # proteusins_cluster_polytheonamide_a.draw_product(
    #     as_string=False, out_file="tailored_test_proteusin_peptide.svg", draw_straightened=False)
    # sliceotide_cluster.make_peptide()
    # sliceotide_cluster.draw_product(as_string=False, out_file = "ripp-test.svg")
    # # print(get_tailoring_sites_atom_names(
    # #     sliceotide_cluster.chain_intermediate))
    # # sliceotide_cluster.draw_product(
    # #     as_string=False, out_file="peptide_test_sliceotide_cluster.svg")
    # sliceotide_cluster.draw_precursor_with_modified_product(as_string=False, out_file="bubbles.svg")
    # sliceotide_cluster.do_tailoring()
    # sliceotide_cluster.draw_product(
    #     as_string=False, out_file="tailored_test_sliceotide_cluster.svg", draw_straightened=False)
    # lasso_peptide_cluster.do_tailoring()
    ## draw_terpene_structure(terpene_cluster)
    # draw_alkaloid_structure(alkaloid_cluster)
    # cyanobactin_cluster_trunkamide.make_peptide()
    # #cyanobactin_cluster_trunkamide.draw_precursor_with_modified_product(as_string=False, out_file= "trunkamide_peptide.svg")
    # cyanobactin_cluster_trunkamide.do_tailoring()
    # cyanobactin_cluster_trunkamide.do_macrocyclization()
    # drawer = Drawer(cyanobactin_cluster_trunkamide.chain_intermediate)
    # # drawer.show_molecule()

    # cyanobactin_cluster_trunkamide.draw_product(as_string=False, out_file= "trunkamide.svg")
    # draw_terpene_structure(terpene_cluster)
    # draw_alkaloid_structure(alkaloid_cluster)

    # draw_cluster(trans_at_ks_cluster_repr, outfile="iterative_pks.svg")



from pikachu.reactions.functional_groups import (
    find_atoms,
    find_bonds,
    combine_structures,
    GroupDefiner,
)
from pikachu.reactions.basic_reactions import internal_condensation
from pikachu.general import read_smiles
from raichu.data.attributes import ATTRIBUTES
from raichu.data.molecular_moieties import PYROPHOSPHATE_BOND
from raichu.reactions.general import initialise_atom_attributes


def reductive_bond_breakage(atom1, atom2, structure):
    """
    Returns the product with a broken bond as a PIKAChU Structure object. 
    If the bond breakage leads to disconnected structures, 
    the structure containing the first atom will be kept.

    structure: PIKAChU Structure object of the intermediate
    atom1: atom on side of the bond to be kept
    atom2. atom on the other side of the bond
    """
    structure = structure.kekulise()
    atom1 = structure.get_atom(atom1)
    atom2 = structure.get_atom(atom2)
    bond = atom1.get_bond(atom2)
    assert bond
    assert bond.type == "single"

    structure.break_bond(bond)
    structure.add_atom("H", [atom1])
    structure.add_atom("H", [atom2])
    initialise_atom_attributes(structure)
    structures = structure.split_disconnected_structures()
    for structure in structures:
        if atom1 in structure.graph:
            structure.refresh_structure(find_cycles=True)
            return structure


def excise_from_structure(atom1, atom2, structure):
    """
    Returns the product without the group between atoms as a PIKAChU Structure object

    structure: PIKAChU Structure object of the intermediate
    atom1: last atom in structure on one side
    atom2. last atom in structure on other side
    """
    # find exact bonds  to the atoms between:

    traveling_atom = atom1
    target = atom2
    visited = []
    dead_ends = []
    atom_before = None
    while traveling_atom != target:
        if all(atom in visited for atom in traveling_atom.neighbours):
            dead_ends.append(traveling_atom)

            traveling_options = [
                atom for atom in traveling_atom.neighbours if atom not in dead_ends
            ]
            if len(traveling_options) > 0:
                traveling_atom = traveling_options[0]
            else:
                raise ValueError(f"Atoms are not connected")
        else:
            neighbour = [
                atom for atom in traveling_atom.neighbours if atom not in visited
            ][0]
            visited.append(neighbour)
            atom_before = traveling_atom
            traveling_atom = neighbour
    atom_before_atom2 = atom_before
    bond_to_break_2 = atom2.get_bond(atom_before_atom2)

    visited = []
    dead_ends = []
    atom_before = None
    traveling_atom = atom2
    target = atom1
    while traveling_atom != target:
        if all(atom in visited for atom in traveling_atom.neighbours):
            dead_ends.append(traveling_atom)

            traveling_options = [
                atom for atom in traveling_atom.neighbours if atom not in dead_ends
            ]
            if len(traveling_options) > 0:
                traveling_atom = traveling_options[0]
            else:
                raise ValueError(f"Atoms are not connected")
        else:
            neighbour = [
                atom for atom in traveling_atom.neighbours if atom not in visited
            ][0]
            visited.append(neighbour)
            atom_before = traveling_atom
            traveling_atom = neighbour
    atom_before_atom1 = atom_before
    bond_to_break_1 = atom1.get_bond(atom_before_atom1)
    structure.break_bond(bond_to_break_1)
    structure.break_bond(bond_to_break_2)
    structure.make_bond(atom1, atom2, structure.find_next_bond_nr())
    structures = structure.split_disconnected_structures()
    for structure in structures:
        if atom1 in structure.graph:
            return structure


def change_chirality(carbon, structure):
    carbon = structure.get_atom(carbon)
    if carbon.chiral == "clockwise":
        carbon.chiral = "counterclockwise"
    elif carbon.chiral == "counterclockwise":
        carbon.chiral = "clockwise"
    initialise_atom_attributes(structure)
    structure.refresh_structure(find_cycles=True)
    return structure


def remove_atom(atom, structure):
    """
    Returns the product without the group attached to the atom as a PIKAChU Structure object

    structure: PIKAChU Structure object of the intermediate
    atom: atom to be removed
    """
    atom = structure.get_atom(atom)
    carbon = atom.get_neighbour("C")
    bond = atom.get_bond(carbon)
    assert bond
    structure.break_bond(bond)
    structure_1, structure_2 = structure.split_disconnected_structures()
    if carbon in structure_1.graph:
        structure = structure_1
    else:
        structure = structure_2
    structure.add_atom("H", [carbon])
    initialise_atom_attributes(structure)
    structure.refresh_structure(find_cycles=True)
    return structure


def remove_group_at_bond(bond, structure):
    """
    Returns the product without the grup attached to the bond as a PIKAChU Structure object

    structure: PIKAChU Structure object of the intermediate
    bond: PIKAChU bond to be broken
    """
    carbon = bond.get_neighbour("C")
    structure.break_bond(bond)
    structure_1, structure_2 = structure.split_disconnected_structures()
    if carbon in structure_1.graph:
        structure = structure_1
    else:
        structure = structure_2
    structure.add_atom("H", [carbon])
    initialise_atom_attributes(structure)
    structure.refresh_structure(find_cycles=True)
    return structure


def dephosphorylation(structure):
    """
    Returns the dephosphorylated product as a PIKAChU Structure object

    structure: PIKAChU Structure object of the intermediate
    """
    pyrophosphate_bonds = find_bonds(PYROPHOSPHATE_BOND, structure)
    if len(pyrophosphate_bonds) > 0:
        pyrophosphate_bond = pyrophosphate_bonds[0]
        carbon = pyrophosphate_bond.get_neighbour("C")
        structure.break_bond(pyrophosphate_bond)
        structure_1, structure_2 = structure.split_disconnected_structures()
        if carbon in structure_1.graph:
            structure = structure_1
        else:
            structure = structure_2
        structure.add_atom("H", [carbon])
        initialise_atom_attributes(structure)
    structure.refresh_structure(find_cycles=True)
    return structure


def double_bond_shift(
    structure,
    old_double_bond_atom1,
    old_double_bond_atom2,
    new_double_bond_atom1,
    new_double_bond_atom2,
):
    """
    Returns the reduced product as a PIKAChU Structure object

    old_double_bond_atom1: C atom1 in old double bond
    old_double_bond_atom2: C atom2 in old double bond
    new_double_bond_atom1: C atom1 in new double bond
    new_double_bond_atom2: C atom2 in new double bond
    structure: PIKAChU Structure object of the intermediate
    """
    new_double_bond = structure.bond_lookup[new_double_bond_atom1][
        new_double_bond_atom2
    ]
    assert new_double_bond
    structure = double_bond_reduction(
        old_double_bond_atom1, old_double_bond_atom2, structure
    )
    structure = single_bond_oxidation(
        new_double_bond_atom1, new_double_bond_atom2, structure
    )

    return structure


def single_bond_oxidation(atom1, atom2, structure):
    """
    Returns the oxidized product as a PIKAChU Structure object

    atom1: C atom1 in single bond
    atom2: C atom2 in single bond
    structure: PIKAChU Structure object of the intermediate
    """

    # remove h atoms
    for selected_atom in [atom1, atom2]:
        bond_to_break = None
        hydrogen = None

        for atom in structure.graph:
            if atom == selected_atom:
                for neighbour in atom.neighbours:
                    if neighbour.type == "H":
                        hydrogen = neighbour
                        for bond in neighbour.bonds:
                            bond_to_break = bond
                        break

        assert bond_to_break and hydrogen
        structure.break_bond(bond_to_break)
    atom1 = structure.get_atom(atom1)
    atom2 = structure.get_atom(atom2)
    single_bond = structure.bond_lookup[atom1][atom2]
    single_bond.make_double()
    structures = structure.split_disconnected_structures()
    for structure in structures:
        if atom1 in structure.graph:
            initialise_atom_attributes(structure)
            structure.refresh_structure(find_cycles=True)
            return structure


def double_bond_reduction(atom1, atom2, structure):
    """
    Returns the reduced product as a PIKAChU Structure object

    atom1: C atom1 in double bond
    atom2: C atom2 in double bond
    structure: PIKAChU Structure object of the intermediate
    """
    double_bond = atom1.get_bond(atom2)
    assert double_bond
    if double_bond.type == "aromatic":
        structure = structure.kekulise()
        atom1 = structure.get_atom(atom1)
        double_bond = atom1.get_bond(atom2)
        double_bond.make_single()
        double_bond.set_bond_summary()
        for neighbour in double_bond.neighbours:
            structure.add_atom("H", [neighbour])
        structure.refresh_structure(find_cycles=True)
        structure.aromatic_cycles = structure.find_aromatic_cycles()
        structure.aromatic_systems = structure.find_aromatic_systems()
    if double_bond.type == "double":
        double_bond.make_single()
        double_bond.set_bond_summary()
        for neighbour in double_bond.neighbours:
            structure.add_atom("H", [neighbour])
    initialise_atom_attributes(structure)
    structure.refresh_structure(find_cycles=True)
    return structure


def epoxidation(atom1, atom2, structure):
    """
    Returns the epoxidated product as a PIKAChU Structure object

    atom1: C atom1 to be epoxidated
    atom2: C atom2 to be epoxidated, needs to be neighbour of atom1
    structure: PIKAChU Structure object of the intermediate
    """
    assert atom1.type == "C"
    assert atom2.type == "C"
    double_bond_reduction(atom1, atom2, structure)
    structure = addition(atom1, "O", structure)
    initialise_atom_attributes(structure)
    structure.refresh_structure()
    # find O atom
    oxygen = atom1.get_neighbour("O")
    structure = oxidative_bond_formation(oxygen, atom2, structure)
    initialise_atom_attributes(structure)
    structure.refresh_structure()
    return structure


def cyclisation(structure, atom1, atom2):
    """Performs cyclisation

    atom1: PIKAChU atom to be used in cyclisation that will remain in final product
    atom2: PIKAChU atom to be used in cyclisation that will be removed from final product -> needs to be oxygen
    structure: PIKAChU Structure object to perform cyclization on
    """
    if atom2.type != "O":
        atom1, atom2 = atom2, atom1
    cyclisation_site_1 = structure.get_atom(atom1)
    h_atom = cyclisation_site_1.get_neighbour("H")
    assert h_atom
    h_bond = cyclisation_site_1.get_bond(h_atom)
    assert h_bond
    cyclisation_site_2 = structure.get_atom(atom2)
    c_atom = cyclisation_site_2.get_neighbour("C")
    assert c_atom
    c_bond = cyclisation_site_2.get_bond(c_atom)
    assert c_bond

    cyclic_product, water = internal_condensation(structure, c_bond, h_bond)

    return cyclic_product


def oxidative_bond_formation(atom1, atom2, structure):
    """Performs cyclisation

    atom1: PIKAChU atom to be used in cyclisation
    atom2: PIKAChU atom to be used in cyclisation
    structure: PIKAChU Structure object to perform cyclization on
    """

    cyclisation_site_1 = structure.get_atom(atom1)
    h_atom = cyclisation_site_1.get_neighbour("H")
    assert h_atom
    h_bond = cyclisation_site_1.get_bond(h_atom)
    assert h_bond
    cyclisation_site_2 = structure.get_atom(atom2)
    h_atom_2 = cyclisation_site_2.get_neighbour("H")
    assert h_atom_2
    h_bond_2 = cyclisation_site_2.get_bond(h_atom_2)
    assert h_bond_2
    structure.break_bond(h_bond)
    structure.break_bond(h_bond_2)

    # Create the bonds
    structure.make_bond(
        cyclisation_site_1, cyclisation_site_2, structure.find_next_bond_nr()
    )
    structure.make_bond(h_atom, h_atom_2, structure.find_next_bond_nr())
    # remove from chirality dict

    cyclisation_site_1 = structure.get_atom(cyclisation_site_1)
    cyclisation_site_2 = structure.get_atom(cyclisation_site_2)

    # update chiral_dict -remove with new pikachu update
    bonds_atom_1 = cyclisation_site_1.bonds[:]
    bonds_atom_2 = cyclisation_site_2.bonds[:]

    for bond in bonds_atom_1:
        if h_atom in bond.chiral_dict:
            bond.chiral_dict[cyclisation_site_2] = bond.chiral_dict[h_atom]
            del bond.chiral_dict[h_atom]
        for atom, atoms_and_chirality in bond.chiral_dict.items():
            if h_atom in atoms_and_chirality:
                atoms_and_chirality[cyclisation_site_2] = atoms_and_chirality[h_atom]
                del atoms_and_chirality[h_atom]

    for bond in bonds_atom_2:
        if h_atom_2 in bond.chiral_dict:
            bond.chiral_dict[cyclisation_site_1] = bond.chiral_dict[h_atom_2]
            del bond.chiral_dict[h_atom_2]
        for atom, atoms_and_chirality in bond.chiral_dict.items():
            if h_atom_2 in atoms_and_chirality:
                atoms_and_chirality[cyclisation_site_1] = atoms_and_chirality[h_atom_2]
                del atoms_and_chirality[h_atom_2]

    # Put the h2 and the product into different Structure instances
    structure.refresh_structure(find_cycles=False)
    structures = structure.split_disconnected_structures()
    product = None
    initialise_atom_attributes(structure)

    # Find out which of the structures is your product and which is your h2

    for structure in structures:
        if cyclisation_site_1 in structure.graph:
            structure.refresh_structure(find_cycles=False)
            initialise_atom_attributes(structure)
            product = structure
    return product


def proteolytic_cleavage(bond, structure, structure_to_keep: str = "end"):
    """Performs proteolytic cleavage

    bond: exact PIKAChU bond object to cleave
    structure: PIKAChU Structure object to perform cleavage on
    structure_to_keep: determines if the leading or the following peptide should be kept ("leader" or "follower")
    """
    carbon = bond.get_neighbour("C")
    assert carbon
    nitrogen = bond.get_neighbour("N")
    assert nitrogen
    structure.break_bond(bond)
    oxygen = structure.add_atom("O", [carbon])
    initialise_atom_attributes(structure)
    structure.add_atom("H", [oxygen])
    structure.add_atom("H", [nitrogen])

    initialise_atom_attributes(structure)

    structures = structure.split_disconnected_structures()
    for structure in structures:
        if structure_to_keep == "start":
            if carbon in structure.graph:
                structure.refresh_structure()
                return structure
        if structure_to_keep == "end":
            if nitrogen in structure.graph:
                structure.refresh_structure()
                return structure


def find_atoms_for_tailoring(chain_intermediate, atom_type):
    """
    Returns atoms that can be tailored

     chain_intermediate: PIKAChU Structure object of a polyketide/NRP
     atom_type: type of atom to search for (e.g. C)
    """

    for atom in chain_intermediate.graph:
        atom.hybridise()
    atoms_filtered = []
    for atom in chain_intermediate.graph:
        if atom.type == atom_type:
            neighbour_types = []
            for neighbour in atom.neighbours:
                neighbour_types.append(neighbour.type)
            if atom not in atoms_filtered and neighbour_types.count("H") >= 1:
                atoms_filtered.append(atom)

    return atoms_filtered


def addition(target_atom, structure_to_add, structure):
    """
    Returns the structure that has a group added at the target atom.

    structure_to_add: SMILES of structure, atom to be added needs to be at first position
    structure: PIKAChU Structure object
    target_atom:  PIKAChU atom object
    """
    FIRST_ATOM = GroupDefiner("first_atom", structure_to_add, 0)
    group_to_add = read_smiles(structure_to_add)
    group_to_add.add_attributes(ATTRIBUTES, boolean=True)
    atom_to_add = find_atoms(FIRST_ATOM, group_to_add)[0]
    target_atom = structure.get_atom(target_atom)
    hydrogen_1 = atom_to_add.get_neighbour("H")
    bond_1 = atom_to_add.get_bond(hydrogen_1)
    hydrogen_2 = target_atom.get_neighbour("H")
    if not hydrogen_2:
        raise Exception("Can't add to this atom!")

    bond_2 = target_atom.get_bond(hydrogen_2)

    combined_structure = combine_structures([group_to_add, structure])
    combined_structure.break_bond(bond_1)
    combined_structure.break_bond(bond_2)

    combined_structure.make_bond(
        atom_to_add, target_atom, combined_structure.find_next_bond_nr()
    )
    combined_structure.make_bond(
        hydrogen_1, hydrogen_2, combined_structure.find_next_bond_nr()
    )
    structures = combined_structure.split_disconnected_structures()
    for structure in structures:
        if atom_to_add.nr in structure.atoms:
            structure.refresh_structure()
            initialise_atom_attributes(structure)
            return structure


def cyclodehydration(structure, atom1, oxygen):
    initialise_atom_attributes(structure)
    carbon = oxygen.get_neighbour("C")
    nitrogen = carbon.get_neighbour("N")
    structure = double_bond_reduction(carbon, oxygen, structure)
    structure = oxidative_bond_formation(carbon, atom1, structure)
    structure = remove_atom(oxygen, structure)
    structure = single_bond_oxidation(carbon, nitrogen, structure)
    structure.refresh_structure()
    initialise_atom_attributes(structure)
    return structure


def macrolactam_formation(structure, o_oh_carboxyl):
    """Performs the macrolactam formation on the input chain_intermediate
    using the -OH group defined by the input O-atom participating in that
    internal -OH group and the N-terminus, returns the circular product as PIKAChU Structure
    object.

    chain_intermediate: PIKAChU Structure object of a polyketide
    o_oh_carboxyl: PIKAChU Atom object of the O-atom in the -OH group that the function should use to perform the
    macrolactame reaction.
    """
    cyclisation_site = structure.get_atom(o_oh_carboxyl)
    c_atom = cyclisation_site.get_neighbour("C")
    assert c_atom
    oh_bond = cyclisation_site.get_bond(c_atom)
    assert oh_bond
    terminal_nitrogen = None
    for atom in structure.graph:
        if (
            atom.annotations.in_central_chain
            and atom.type == "N"
            and [neighbour.type for neighbour in atom.neighbours].count("H") == 2
        ):
            terminal_nitrogen = atom
    assert terminal_nitrogen
    h_atom = terminal_nitrogen.get_neighbour("H")
    h_bond = terminal_nitrogen.get_bond(h_atom)
    cyclic_product, water = internal_condensation(structure, oh_bond, h_bond)
    cyclic_product.refresh_structure()
    initialise_atom_attributes(cyclic_product)
    return cyclic_product



from pikachu.drawing.drawing import Drawer

from raichu.drawing.drawer import RaichuDrawer
from raichu.drawing.pathway import draw_arrow_and_text, draw_double_arrow
from raichu.data.attributes import ENZYME_TO_NAME
from raichu.tailoring_enzymes import TailoringEnzyme
from raichu.reactions.general_tailoring_reactions import cyclisation, oxidative_bond_formation


class Cluster:
    def __init__(self, tailoring_representations=None, macrocyclisation_representations=None) -> None:
        self.tailoring_representations = tailoring_representations
        self.macrocyclisation_representations = macrocyclisation_representations

        self.ripp = False
        self.terpene = False
        self.modular = False

        self.chain_intermediate = None
        self.tailored_product = None
        self.cyclase_name = None

        self.linear_product = None
        self.cyclic_product = None
        self.cyclic_products = []

        self.modular_intermediates = []
        self.cyclic_intermediates = []
        self.tailoring_intermediates = []
        self.cleaved_intermediates = []

    def initialize_macrocyclization(self):
        initialized_macrocyclization_atoms = []
        if self.macrocyclisation_representations:
            for cyclization in self.macrocyclisation_representations:
                atom_1 = None
                atom_2 = None
                for atom in self.chain_intermediate.atoms.values():
                    if str(atom) == cyclization.atom_1:
                        atom_1 = atom
                    elif str(atom) == cyclization.atom_2:
                        atom_2 = atom
                atoms = [atom_1, atom_2]
                # atoms = [atom for atom in self.chain_intermediate.atoms.values()
                #          if str(atom) in [cyclization.atom_1, cyclization.atom_2]]
                initialized_macrocyclization_atoms.append((atoms, cyclization.type))
        return initialized_macrocyclization_atoms

    def do_macrocyclization(self, sequential=True):
        initialized_macrocyclization_atoms = self.initialize_macrocyclization()
        self.cyclic_intermediates.append(self.chain_intermediate.deepcopy())

        for index, macrocyclization_atoms in enumerate(initialized_macrocyclization_atoms):
            atoms, cyclisation_type = macrocyclization_atoms
            if [str(atom) for atom in atoms] != [self.macrocyclisation_representations[index].atom_1,
                                                 self.macrocyclisation_representations[index].atom_2]:
                raise ValueError(
                    f'Not all atoms {str(initialized_macrocyclization_atoms)} \
                    for macrocyclisation exist in the structure.')

            if sequential:
                structure = self.chain_intermediate
            else:
                structure = self.chain_intermediate.deepcopy()

            atom1 = structure.get_atom(atoms[0])
            atom2 = structure.get_atom(atoms[1])

            if cyclisation_type == 'oxidative':
                structure = oxidative_bond_formation(atom1, atom2, structure)

            elif cyclisation_type == 'condensative':
                structure = cyclisation(structure, atom1, atom2)
            else:
                raise ValueError("Cyclisation type must be oxidative or condensative.")

            if sequential:
                self.chain_intermediate = structure
                self.cyclic_intermediates.append(structure.deepcopy())
            else:
                self.cyclic_products.append(structure.deepcopy())

            self.cyclic_product = structure

    def initialize_modification_sites(self, modification_sites):
        modification_sites_initialized = []
        for atom_representations_for_reaction in modification_sites:
            atom_indices_and_types = map(
                lambda atom: [int(''.join(filter(str.isdigit, atom))), atom], atom_representations_for_reaction)
            atoms_in_structure = list(map(
                str, self.chain_intermediate.atoms.values()))
            atoms_for_reaction_initialized = [self.chain_intermediate.atoms[atom[0]]
                                              for atom in atom_indices_and_types if atom[1] in atoms_in_structure]
            atoms_for_reaction_initialized = list(
                filter(lambda atom: atom is not None, atoms_for_reaction_initialized))
            modification_sites_initialized += [
                atoms_for_reaction_initialized]
        return modification_sites_initialized

    def do_tailoring(self):
        if self.tailoring_representations:
            self.tailoring_intermediates.append(self.chain_intermediate.deepcopy())

            for tailoring_enzyme_representation in self.tailoring_representations:
                modification_sites = self.initialize_modification_sites(
                    tailoring_enzyme_representation.modification_sites)

                if [[str(atom) for atom in atoms_for_reaction] for atoms_for_reaction in modification_sites] != tailoring_enzyme_representation.modification_sites:
                    raise ValueError(
                        f'Not all atoms {tailoring_enzyme_representation.modification_sites} for \
                        {tailoring_enzyme_representation.type} exist in the structure.')

                tailoring_enzyme = TailoringEnzyme(tailoring_enzyme_representation.gene_name,
                                                   tailoring_enzyme_representation.type, modification_sites,
                                                   tailoring_enzyme_representation.substrate)
                self.tailored_product = tailoring_enzyme.do_tailoring(self.chain_intermediate)
                self.tailoring_intermediates.append(self.tailored_product.deepcopy())
                self.chain_intermediate = self.tailored_product

    def get_drawings(self, whitespace=30, reaction_type="tailoring"):

        drawings = []
        widths = []
        centre_points = []
        max_height = -1000000000

        if reaction_type == 'tailoring':
            structures = self.tailoring_intermediates
        elif reaction_type == 'cyclisation':
            structures = self.cyclic_intermediates
        elif reaction_type == 'cleavage':
            structures = self.cleaved_intermediates
        else:
            raise ValueError(f"Expected 'tailoring' or 'cyclisation'. Got {reaction_type}.")

        for i, structure in enumerate(structures):

            drawing = Drawer(structure)

            drawing.flip_y_axis()
            drawing.move_to_positive_coords()
            drawing.convert_to_int()

            min_x = 100000000
            max_x = -100000000
            min_y = 100000000
            max_y = -100000000

            for atom in drawing.structure.graph:
                if atom.draw.positioned:
                    if atom.draw.position.x < min_x:
                        min_x = atom.draw.position.x
                    if atom.draw.position.x > max_x:
                        max_x = atom.draw.position.x
                    if atom.draw.position.y < min_y:
                        min_y = atom.draw.position.y
                    if atom.draw.position.y > max_y:
                        max_y = atom.draw.position.y

            width = max_x - min_x + whitespace
            height = max_y - min_y + whitespace
            if height > max_height:
                max_height = height

            widths.append(width)
            centre_points.append(((max_x + min_x) / 2, (max_y + min_y) / 2))
            drawings.append(drawing)

        return drawings, widths, max_height, centre_points

    def draw_pathway(self, min_arrow_size=40, order=("tailoring", "cyclisation"), mode='gene_name', as_string=False,
                     out_file=None, summary=False):

        all_drawings = []
        all_widths = []
        max_height = 0
        all_centre_points = []
        enzyme_names = []

        for i, reaction_type in enumerate(order):
            if reaction_type == "tailoring":
                assert self.tailoring_representations
                assert mode in ['gene_name', 'reaction_name']
                drawings, widths, height, centre_points = self.get_drawings(reaction_type='tailoring')
                for enzyme in self.tailoring_representations:
                    if mode == 'gene_name':
                        text = enzyme.gene_name
                    elif mode == 'reaction_name':
                        text = ENZYME_TO_NAME[enzyme.type]
                    else:
                        raise ValueError("Drawing mode must be 'gene_name' or 'reaction_name'.")
                    enzyme_names.append(text)

            elif reaction_type == "cyclisation":
                assert self.macrocyclisation_representations
                drawings, widths, height, centre_points = self.get_drawings(reaction_type='cyclisation')
                for _ in self.macrocyclisation_representations:
                    if mode == 'gene_name':
                        if self.cyclase_name:
                            text = self.cyclase_name
                        else:
                            text = ''
                    elif mode == 'reaction_name':
                        text = 'cyclisation'
                    else:
                        raise ValueError("Drawing mode must be 'gene_name' or 'reaction_name'.")
                    enzyme_names.append(text)
            elif reaction_type == "cleavage":
                assert self.cleaved_intermediates
                drawings, widths, height, centre_points = self.get_drawings(reaction_type='cleavage')
                for _ in range(len(self.cleaved_intermediates) - 1):
                    text = 'proteolytic cleavage'
                    enzyme_names.append(text)
            else:
                raise ValueError(f"Expected 'tailoring' or 'cyclisation'. Got {reaction_type}.")

            assert len(drawings) > 1

            if i != 0:
                drawings = drawings[1:]
                widths = widths[1:]
                centre_points = centre_points[1:]

            all_drawings += drawings
            all_widths += widths
            all_centre_points += centre_points
            if height > max_height:
                max_height = height

        target_x = 0
        target_y = max_height / 2
        arrow_svgs = []
        structure_svgs = []
        padding = None
        svg_style = None

        if summary:
            all_drawings = [all_drawings[0], all_drawings[-1]]
            all_widths = [all_widths[0], all_widths[-1]]
            all_centre_points = [all_centre_points[0], all_centre_points[-1]]

        for i, drawing in enumerate(all_drawings):

            current_x, current_y = all_centre_points[i]
            width = all_widths[i]
            translation_x = (target_x + 0.5 * width) - current_x
            translation_y = target_y - current_y

            drawing.set_structure_id(f"s{i}")
            svg_style = drawing.svg_style

            target_x += width
            if not summary:
                if i != len(all_drawings) - 1:
                    enzyme_name = enzyme_names[i]
                    arrow_size = max(len(enzyme_name) * 9, min_arrow_size)

                    arrow_start = target_x
                    arrow_end = target_x + arrow_size
                    arrow_y = target_y

                    arrow_svg = draw_arrow_and_text(arrow_start, arrow_end, arrow_y, i, enzyme_name)
                    arrow_svgs.append(arrow_svg)

                    target_x += arrow_size

            elif i != len(all_drawings) - 1:
                arrow_start = target_x
                arrow_end = target_x + min_arrow_size
                arrow_y = target_y
                arrow_svg = draw_double_arrow(arrow_start, arrow_end, arrow_y, i)
                arrow_svgs.append(arrow_svg)
                target_x += min_arrow_size

            drawing.move_structure(translation_x, translation_y)
            structure_svg = drawing.draw_svg()
            structure_svgs.append(structure_svg)
            padding = drawing.options.padding

        assert padding

        x1 = 0
        x2 = target_x + padding
        y1 = 0
        y2 = max_height + padding

        width = x2
        height = y2

        svg_string = f"""<svg width="{width}" height="{height}" viewBox="{x1} {y1} {x2} {y2}" xmlns="http://www.w3.org/2000/svg">\n"""
        if svg_style:
            svg_string += f"{svg_style}\n"

        svg_string += f"""<g id="arrows">\n"""
        for arrow in arrow_svgs:
            svg_string += arrow
        svg_string += "</g>\n"

        for string in structure_svgs:
            svg_string += string

        svg_string += "</svg>"

        if as_string:
            return svg_string
        else:
            if out_file is None:
                raise ValueError("Must provide output svg path if 'as_string' is set to False.")
            else:
                with open(out_file, 'w') as svg_out:
                    svg_out.write(svg_string)

    def draw_tailoring(self, arrow_size=40, mode='gene_name', as_string=False, out_file=None):

        assert self.tailoring_representations
        assert mode in ['gene_name', 'reaction_name']
        drawings, widths, max_height, centre_points = self.get_drawings()

        target_x = 0
        target_y = max_height / 2
        arrow_svgs = []
        structure_svgs = []
        padding = None
        svg_style = None

        for i, drawing in enumerate(drawings):

            current_x, current_y = centre_points[i]
            width = widths[i]
            translation_x = (target_x + 0.5 * width) - current_x
            translation_y = target_y - current_y

            drawing.set_structure_id(f"s{i}")
            svg_style = drawing.svg_style

            target_x += width
            if i != len(drawings) - 1:
                enzyme = self.tailoring_representations[i]

                arrow_start = target_x
                arrow_end = target_x + arrow_size
                arrow_y = target_y
                if mode == 'gene_name':
                    text = enzyme.gene_name
                elif mode == 'reaction_name':
                    text = ENZYME_TO_NAME[enzyme.type]
                else:
                    raise ValueError("Drawing mode must be 'gene_name' or 'reaction_name'.")

                arrow_svg = draw_arrow_and_text(arrow_start, arrow_end, arrow_y, i, text)
                arrow_svgs.append(arrow_svg)

            target_x += arrow_size

            drawing.move_structure(translation_x, translation_y)
            structure_svg = drawing.draw_svg()
            structure_svgs.append(structure_svg)
            padding = drawing.options.padding

        assert padding

        x1 = 0
        x2 = target_x + padding
        y1 = 0
        y2 = max_height + padding

        width = x2
        height = y2

        svg_string = f"""<svg width="{width}" height="{height}" viewBox="{x1} {y1} {x2} {y2}" xmlns="http://www.w3.org/2000/svg">\n"""
        if svg_style:
            svg_string += f"{svg_style}\n"

        svg_string += f"""<g id="arrows">\n"""
        for arrow in arrow_svgs:
            svg_string += arrow
        svg_string += "</g>\n"

        for string in structure_svgs:
            svg_string += string

        svg_string += "</svg>"

        if as_string:
            return svg_string
        else:
            if out_file is None:
                raise ValueError("Must provide output svg path if 'as_string' is set to False.")
            else:
                with open(out_file, 'w') as svg_out:
                    svg_out.write(svg_string)

    def draw_product(self, as_string=False, out_file=None, draw_Cs_in_pink = False):
        assert self.chain_intermediate
        drawing = RaichuDrawer(self.chain_intermediate, dont_show=True, add_url=True, draw_Cs_in_pink=draw_Cs_in_pink,
                               make_linear=False)
        drawing.draw_structure()
        svg_string = drawing.get_svg_string_matplotlib()

        if as_string:
            return svg_string
        else:
            if out_file is None:
                raise ValueError("Must provide output svg directory if 'as_string' is set to False.")
            else:
                with open(out_file, 'w') as svg_out:
                    svg_out.write(svg_string)



from typing import List

from pikachu.general import read_smiles

from raichu.substrate import TerpeneCyclaseSubstrate
from raichu.representations import (
    MacrocyclizationRepresentation,
    TailoringRepresentation,
    IsomerizationRepresentation,
    MethylShiftRepresentation,
)
from raichu.reactions.general_tailoring_reactions import (
    dephosphorylation,
    oxidative_bond_formation,
    double_bond_reduction,
)
from raichu.cluster.base_cluster import Cluster
from raichu.reactions.general_tailoring_reactions import (
    double_bond_shift,
    reductive_bond_breakage,
    addition
)


class TerpeneCluster(Cluster):
    def __init__(
        self,
        cyclase_name: str,
        precursor: str,
        cyclase_type: str = None,
        macrocyclisations: List[MacrocyclizationRepresentation] = None,
        double_bond_isomerisations: List[IsomerizationRepresentation] = None,
        methyl_shifts: List[MethylShiftRepresentation] = None,
        tailoring_representations: List[TailoringRepresentation] = None,
    ) -> None:
        super().__init__(tailoring_representations, macrocyclisations)
        self.cyclase_name = cyclase_name
        self.cyclase_type = cyclase_type
        self.precursor = precursor
        self.isomerization_representations = double_bond_isomerisations
        self.methyl_shift_representations = methyl_shifts

        self.chain_intermediate = None
        self.tailored_product = None
        self.final_product = None
        self.terpene = True

    def create_precursor(self) -> None:
        substrate = TerpeneCyclaseSubstrate(self.precursor)
        self.chain_intermediate = read_smiles(substrate.smiles)

    def do_double_bond_isomerization(self):
        if not self.isomerization_representations:
            return
        initialized_isomerization_atoms = self.initialize_modification_sites(
            [
                isomerization_representation.modification_sites
                for isomerization_representation in self.isomerization_representations
            ]
        )
        for atoms in initialized_isomerization_atoms:
            if len(atoms) < 4:
                continue
            old_double_bond_atom1 = self.chain_intermediate.get_atom(atoms[0])
            old_double_bond_atom2 = self.chain_intermediate.get_atom(atoms[1])
            new_double_bond_atom1 = self.chain_intermediate.get_atom(atoms[2])
            new_double_bond_atom2 = self.chain_intermediate.get_atom(atoms[3])
            if len(set(atoms)) == len(atoms):
                raise ValueError(
                    "The bonds need to be adjacent to perform a double bond shift."
                )
            self.chain_intermediate = double_bond_shift(
                self.chain_intermediate,
                old_double_bond_atom1,
                old_double_bond_atom2,
                new_double_bond_atom1,
                new_double_bond_atom2,
            )


    def do_methyl_shift(self):
        if not self.methyl_shift_representations:
            return
        initialized_methyl_shift_atoms = self.initialize_modification_sites(
                [
                    methyl_shift_representation.modification_sites
                    for methyl_shift_representation in self.methyl_shift_representations
                ]
            )
        for atoms in initialized_methyl_shift_atoms:
                if len(atoms) != 2:
                    continue
                transferred_c = atoms[0]
                # Assert its actually a methyl group
                assert [atom.type for atom in transferred_c.neighbours].count("H") == 3
                source = None
                source = [
                    atom for atom in transferred_c.neighbours if atom.type != "H"
                ][0]
                assert source
                destination_c = atoms[1]
                assert destination_c.has_neighbour("H")
                assert transferred_c.type == "C" and destination_c.type == "C"

                self.chain_intermediate = reductive_bond_breakage(source, transferred_c, self.chain_intermediate)
                self.chain_intermediate = addition(destination_c, "C", self.chain_intermediate)
                self.chain_intermediate.refresh_structure(find_cycles=True)



    def do_macrocyclization(self, sequential=True):
        initialized_macrocyclization_atoms = self.initialize_macrocyclization()
        self.cyclic_intermediates.append(self.chain_intermediate.deepcopy())
        self.do_double_bond_isomerization()
        self.do_methyl_shift()
        if self.cyclase_type != "Class_2":
            self.chain_intermediate = dephosphorylation(self.chain_intermediate)

        for (
            macrocyclization_atoms,
            cyclisation_type,
        ) in initialized_macrocyclization_atoms:
            atom1 = self.chain_intermediate.get_atom(macrocyclization_atoms[0])
            atom2 = self.chain_intermediate.get_atom(macrocyclization_atoms[1])
            found_bond = False
            for atom in [atom1, atom2]:
                if found_bond and any(
                    [neighbour.type == "H" for neighbour in atom.neighbours]
                ):
                    break
                for bond in atom.bonds:
                    if bond.type == "double":
                        self.chain_intermediate = double_bond_reduction(
                            *bond.neighbours, self.chain_intermediate
                        )
                        found_bond = True
                        break
            self.chain_intermediate = oxidative_bond_formation(
                atom1, atom2, self.chain_intermediate
            )
            self.cyclic_intermediates.append(self.chain_intermediate.deepcopy())

        self.cyclic_product = self.chain_intermediate



import pygame

from interactive.domain import Domain
from interactive.style import DOMAIN_SIZE, GENE_LABEL_SIZE, MODULE_SPACING, MODULE_COLOUR, \
    MODULE_LINE_COLOUR, MODULE_PADDING, MODULE_HIGHLIGHT_COLOUR


class Module:
    module_to_order = {'NRPS': ['C', 'A', 'nMT', 'PCP', 'E', 'TE'],
                       'PKS': ['KS', 'AT', 'DH', 'ER', 'KR', 'ACP', 'TE']}

    domain_to_subsequent_domains = {'C': ['A', 'nMT', 'PCP', 'E', 'TE'],
                                    'A': ['nMT', 'PCP', 'E', 'TE'],
                                    'nMT': ['PCP', 'E', 'TE'],
                                    'PCP': ['E', 'TE'],
                                    'E': ['TE'],
                                    'KS': ['AT', 'DH', 'ER', 'KR', 'ACP', 'TE'],
                                    'AT': ['DH', 'ER', 'KR', 'ACP', 'TE'],
                                    'DH': ['ER', 'KR', 'ACP', 'TE'],
                                    'ER': ['KR', 'ACP', 'TE'],
                                    'KR': ['ACP', 'TE'],
                                    'ACP': [],
                                    'TE': []}

    def __init__(self, screen, module_index, module_type, gene):
        assert module_type in ['NRPS', 'PKS']

        self.screen = screen
        self.id = module_index
        self.gene = gene
        self.type = module_type
        self.domains = []
        self.rectangle = None
        self.dna_coords = None

        self.x = 0
        self.y = 0

        self.width = 0
        self.height = DOMAIN_SIZE

        self.set_rectangle()

        self.selected = False

    def set_rectangle(self):
        self.y = self.gene.y - 5
        self.x = self.gene.x + GENE_LABEL_SIZE
        for module in self.gene.modules:
            if module.id != self.id:
                self.x += len(module.domains) * DOMAIN_SIZE
                self.x += MODULE_SPACING
                self.x += MODULE_PADDING
            else:
                break
        self.x += MODULE_PADDING

        self.width = len(self.domains) * DOMAIN_SIZE + MODULE_SPACING
        self.rectangle = pygame.Rect(self.x, self.y, self.width, self.height)

    def adjust_domain_indices(self, insertion_point):
        for domain in self.domains:
            if domain.id >= insertion_point:
                domain.id += 1
                domain.set_rectangle()

    def draw(self, mouse):
        self.set_rectangle()

        colour = MODULE_COLOUR
        if self.rectangle.collidepoint(mouse) or self.selected:
            for domain in self.domains:
                if domain.rectangle.collidepoint(mouse) or domain.selected:
                    break
            colour = MODULE_HIGHLIGHT_COLOUR

        pygame.draw.rect(self.screen, colour, self.rectangle)
        pygame.draw.rect(self.screen, MODULE_LINE_COLOUR, self.rectangle, 1)

        for domain in self.domains:
            domain.draw(mouse)

    def remove_domain(self, domain):
        domain_index = domain.id
        self.domains.pop(domain_index)

        for domain in self.domains:
            if domain.id > domain_index:
                domain.id -= 1

    def add_domain(self, domain_type, domain_label=None):
        assert domain_type not in [domain.type for domain in self.domains]
        assert domain_type in self.module_to_order[self.type]

        insertion_point = 0

        for i, domain in enumerate(self.domains):
            if domain.type not in self.domain_to_subsequent_domains[domain_type]:
                insertion_point += 1
            else:
                break

        self.adjust_domain_indices(insertion_point)

        domain = Domain(self.screen, domain_type, self, insertion_point, domain_label=domain_label)

        self.domains.insert(insertion_point, domain)












from raichu.cluster.modular_cluster import ModularCluster
import os
from raichu.cluster.ripp_cluster import RiPPCluster
from raichu.cluster.terpene_cluster import TerpeneCluster
from raichu.run_raichu import make_domain, build_cluster
from raichu.domain.domain import TailoringDomain, CarrierDomain, SynthesisDomain, RecognitionDomain, \
    TerminationDomain, UnknownDomain, Domain
from raichu.module import PKSModuleSubtype, NRPSModule, LinearPKSModule, IterativePKSModule, TransATPKSModule,\
    ModuleType
from raichu.domain.domain_types import TailoringDomainType, TerminationDomainType, CarrierDomainType, \
    SynthesisDomainType, RecognitionDomainType
from raichu.cluster.alkaloid_cluster import AlkaloidCluster
from raichu.tailoring_enzymes import TailoringEnzyme, TailoringEnzymeType
from raichu.representations import ClusterRepresentation
from raichu.antismash import load_antismash_gbk, parse_antismash_to_cluster_file


def draw_cluster_from_antiSMASH_output(antismash_file: str, out_file: str) -> None:
    draw_cluster_from_modular_cluster_representation(
        load_antismash_gbk(antismash_file),
        out_file=out_file,
    )


def draw_cluster_from_modular_cluster_representation(cluster_repr: ClusterRepresentation, out_file=None, colour_by_module=True) -> None:
    cluster = build_cluster(cluster_repr, strict=False)
    cluster.compute_structures(compute_cyclic_products=False)
    cluster.do_tailoring()

    if out_file:
        return cluster.draw_cluster(as_string=False, out_file=out_file, colour_by_module=colour_by_module)
    else:
        return cluster.draw_cluster()


def draw_products_from_modular_cluster_representation(cluster_repr: ClusterRepresentation, out_dir) -> None:
    cluster = build_cluster(cluster_repr, strict=False)
    cluster.compute_structures(compute_cyclic_products=False)
    cluster.do_tailoring()
    cluster.cyclise_all()
    cluster.draw_all_products(out_dir)


def draw_product_from_modular_cluster_representation(cluster_repr: ClusterRepresentation, out_file) -> None:
    cluster = build_cluster(cluster_repr)
    cluster.compute_structures(compute_cyclic_products=True)
    cluster.do_tailoring()
    cluster.draw_product(as_string=False, out_file=out_file)


def draw_ripp_structure_from_ripp_cluster(ripp_cluster: RiPPCluster, out_folder: str) -> None:
    if not os.path.exists(out_folder):
        os.mkdir(out_folder)

    ripp_cluster.make_peptide()
    order = []

    ripp_cluster.do_tailoring()
    if ripp_cluster.tailoring_representations:
        ripp_cluster.draw_tailoring(out_file=os.path.join(out_folder, "ripp_tailoring_steps.svg"))
        ripp_cluster.draw_product(
            as_string=False, out_file=os.path.join(out_folder, "ripp_tailoring.svg"))
        order.append("tailoring")
    if ripp_cluster.macrocyclisation_representations:
        ripp_cluster.do_macrocyclization()
        ripp_cluster.draw_product(
            as_string=False, out_file=os.path.join(out_folder, "ripp_macrocyclisation.svg"))
        order.append('cyclisation')

    if ripp_cluster.cleavage_sites:
        ripp_cluster.do_proteolytic_cleavage()
        ripp_cluster.draw_product(
            as_string=False, out_file=os.path.join(out_folder, "cleavage_test_ripp.svg"))
        order.append('cleavage')
    ripp_cluster.draw_product(
        as_string=False, out_file=os.path.join(out_folder, "peptide_test_ripp.svg"))
    ripp_cluster.draw_cluster(as_string=False, out_file=os.path.join(out_folder, 'ripp_inline.svg'))
    ripp_cluster.draw_pathway(out_file=os.path.join(out_folder, "ripp_pathway.svg"), order=tuple(order))


def draw_terpene_structure_from_terpene_cluster(terpene_cluster: TerpeneCluster, out_folder: str) -> None:
    terpene_cluster.create_precursor()
    terpene_cluster.draw_product(
        as_string=False, out_file=os.path.join(out_folder, "precursor_test_terpene.svg"))
    terpene_cluster.do_macrocyclization()
    terpene_cluster.draw_product(
        as_string=False, out_file=os.path.join(out_folder, "macroyclisation_test_terpene.svg"))
    terpene_cluster.do_tailoring()
    terpene_cluster.draw_product(
        as_string=False, out_file=os.path.join(out_folder, "tailoring_test_terpene.svg"))


def draw_alkaloid_structure_from_alkaloid_cluster(alkaloid_cluster: AlkaloidCluster, out_folder: str) -> None:
    alkaloid_cluster.make_scaffold()
    alkaloid_cluster.draw_product(
        as_string=False, out_file=os.path.join(out_folder, "precursor_test_alkaloid.svg"))
    alkaloid_cluster.do_tailoring()
    alkaloid_cluster.draw_product(
        as_string=False, out_file=os.path.join(out_folder, "tailoring_test_alkaloid.svg"))


def get_tailoring_sites(structure, enzyme_name='all', out_file=None):
    tailoring_sites = {}
    if enzyme_name == 'all':
        for enzyme_type in TailoringEnzymeType:
            tailoring_enzyme = TailoringEnzyme("gene", enzyme_type.name)
            tailoring_sites[enzyme_type.name] = tailoring_enzyme.get_possible_sites(
                structure)
    else:
        for enzyme_type in TailoringEnzymeType:
            if enzyme_type.name == enzyme_name:
                tailoring_enzyme = TailoringEnzyme("gene", enzyme_type.name)
                tailoring_sites[enzyme_type.name] = tailoring_enzyme.get_possible_sites(structure, out_file=out_file)
    return tailoring_sites


def get_tailoring_sites_atom_names(structure):
    tailoring_sites = {}
    for enzyme_type in TailoringEnzymeType:
        tailoring_enzyme = TailoringEnzyme("gene", enzyme_type.name)
        tailoring_sites[enzyme_type.name] = [str(atom) if type(atom) != list else [str(subatom) for subatom in atom] for atom in tailoring_enzyme.get_possible_sites(
            structure)]
    return tailoring_sites



from raichu.drawing.colours import AMINO_ACID_FILL_COLOURS, AMINO_ACID_OUTLINE_COLOURS


def make_circle(x_coord, y_coord, size, amino_acid):
    """Easy function to draw circle for the domain visualization. Returns
    matplotlib.patches.circle object
    x_coord: int, x-coordinate of the center of the circle to be drawn
    amino_acid : amino acid to be visualized
    size: radius of circle
    """
    colour = AMINO_ACID_FILL_COLOURS[amino_acid]
    outline_colour = AMINO_ACID_OUTLINE_COLOURS[amino_acid]
    circle = f"""<circle r="{size}" cx="{int(x_coord)}" cy="{int(y_coord)}" stroke="{outline_colour}"
        fill="{colour}" stroke-width="1"/>"""
    return circle


from pikachu.general import read_smiles

from raichu.substrate import AminoAcidSubstrate
from raichu.tailoring_enzymes import TailoringEnzyme
from raichu.drawing.drawer import RaichuDrawer
from raichu.cluster.base_cluster import Cluster


class AlkaloidCluster(Cluster):
    def __init__(self, start_amino_acid: str, tailoring_representations=None) -> None:
        super().__init__(tailoring_representations)
        self.start_amino_acid = start_amino_acid
        self.linear_product = None
        self.final_product = None

    def make_scaffold(self):
        substrate = AminoAcidSubstrate(self.start_amino_acid)
        self.linear_product = read_smiles(substrate.smiles)
        self.chain_intermediate = self.linear_product





from pikachu.general import read_smiles

from raichu.substrate import AminoAcidSubstrate
from raichu.tailoring_enzymes import TailoringEnzyme
from raichu.drawing.drawer import RaichuDrawer

class Alkaloid_Cluster:
    def __init__(self, start_amino_acid: str, tailoring_enzymes_representation = None) -> None:
        self.start_amino_acid = start_amino_acid
        self.tailoring_enzymes_representation = tailoring_enzymes_representation
        self.chain_intermediate = None
        self.linear_product = None
        self.tailored_product = None
        self.final_product = None

    def make_scaffold(self):
        substrate = AminoAcidSubstrate(self.start_amino_acid)
        self.linear_product = read_smiles(substrate.smiles)
        self.chain_intermediate = self.linear_product

    def initialize_modification_sites_on_structure(self, modification_sites):
            modification_sites_initialized = []
            for atoms_for_reaction in modification_sites:
                atoms_for_reaction_with_numbers = map(
                    lambda atom: [int(''.join(filter(str.isdigit, atom))), atom], atoms_for_reaction)
                atoms_in_structure = list(map(
                    str, self.chain_intermediate.atoms.values()))
                atoms_for_reaction_initialized = [self.chain_intermediate.atoms[atom[0]]
                                                  for atom in atoms_for_reaction_with_numbers if atom[1] in atoms_in_structure]
                atoms_for_reaction_initialized = list(
                    filter(lambda atom: atom is not None, atoms_for_reaction_initialized))
                modification_sites_initialized += [
                    atoms_for_reaction_initialized]
            return modification_sites_initialized

    def do_tailoring(self):
        if self.tailoring_enzymes_representation:
            for tailoring_enzyme_representation in self.tailoring_enzymes_representation:
                modification_sites = self.initialize_modification_sites_on_structure(
                    tailoring_enzyme_representation.modification_sites)
                if [[str(atom) for atom in atoms_for_reaction]for atoms_for_reaction in modification_sites] != tailoring_enzyme_representation.modification_sites:
                    raise ValueError(
                        f'Not all atoms {tailoring_enzyme_representation.modification_sites} for {tailoring_enzyme_representation.type} exist in the structure.')
                tailoring_enzyme = TailoringEnzyme(
                    tailoring_enzyme_representation.gene_name, tailoring_enzyme_representation.type, modification_sites, tailoring_enzyme_representation.substrate)
                self.tailored_product = tailoring_enzyme.do_tailoring(
                    self.chain_intermediate)
                self.chain_intermediate = self.tailored_product

    def draw_product(self, as_string=True, out_file=None):
            assert self.chain_intermediate
            for atom in self.chain_intermediate.graph:
                if atom.chiral:
                    atom.chiral = False
            drawing = RaichuDrawer(self.chain_intermediate, dont_show=True, add_url=True, draw_Cs_in_pink=False, draw_straightened=False)
            drawing.draw_structure()
            svg_string = drawing.save_svg_string()
            if as_string:
                return svg_string
            else:
                if out_file is None:
                    raise ValueError("Must provide output svg directory if 'as_string' is set to False.")
                else:
                    with open(out_file, 'w') as svg_out:
                        svg_out.write(svg_string)




from pikachu.reactions.functional_groups import find_atoms, find_bonds
from pikachu.reactions.basic_reactions import condensation
from raichu.central_chain_detection.label_central_chain import label_nrp_central_chain
from raichu.reactions.general import label_rest_groups, initialise_atom_attributes, reset_nrp_annotations
from raichu.data.molecular_moieties import C_TERMINAL_OH, N_AMINO_ACID, B_N_AMINO_ACID

def ribosomal_elongation(amino_acid, chain_intermediate, amino_acid_number=None):
    """
    Returns the ribosomal condensation product as a PIKAChU Structure object
    
    amino_acid_number: Type and number of amino acid in peptide (e.g. A14)
    amino_acid: PIKAChU Structure object of the amino acid
    chain_intermediate: PIKAChU Structure object of the ripp intermediate
    """

    # Initialize annotations in all atoms in amino acid
    initialise_atom_attributes(amino_acid)
    for atom in amino_acid.atoms.values():
        atom.amino_acid = amino_acid_number
    amino_acid.atoms
    oh_bond = find_bonds(C_TERMINAL_OH, chain_intermediate)
    assert len(oh_bond)==1
    oh_bond = oh_bond[0]

    # Label rest groups
    label_rest_groups(amino_acid, chain_intermediate)


    # Find amino group of amino acid or beta-amino acid
    n_atoms_aa = find_atoms(N_AMINO_ACID, amino_acid)
    if not n_atoms_aa:
        n_atoms_aa = find_atoms(B_N_AMINO_ACID, amino_acid)

    assert len(n_atoms_aa) == 1

    n_atom = n_atoms_aa[0]
    h_bond = None

    for bond in n_atom.bonds:
        for neighbour in bond.neighbours:
            if neighbour.type == 'H':
                h_bond = bond
                break

    assert h_bond

    # Label atoms in amino acid that end up in central peptide chain
    label_nrp_central_chain(amino_acid)

    # Reset chiral_c_ep and n_atom_nmeth AtomAnnotations attribute for all atoms in the intermediate;
    # This way, only the newly incorporated amino acid is labelled for tailoring
    reset_nrp_annotations(chain_intermediate)

    # Carry out condensation reaction using build-in PIKAChU function
    condensation_product = condensation(chain_intermediate, amino_acid, oh_bond, h_bond)[0]

    # Refresh condensation product
    condensation_product.refresh_structure(find_cycles=True)

    # Initialize annotations for atoms that don't have any yet
    initialise_atom_attributes(condensation_product)

    return condensation_product


from typing import List, Union

from pikachu.general import read_smiles
from pikachu.chem.structure import Structure
from pikachu.chem.atom import Atom

from raichu.class_domain import CarrierDomain as ClassCarrierDomain
from raichu.domain.domain import (
    Domain,
    TailoringDomain,
    RecognitionDomain,
    SynthesisDomain,
    CarrierDomain,
    TerminationDomain,
)
from raichu.central_chain_detection.label_central_chain import (
    label_pk_central_chain,
    label_nrp_central_chain,
)
from raichu.attach_to_domain import attach_to_domain_pk, attach_to_domain_nrp
from enum import Enum, unique
from raichu.substrate import PKSSubstrate


@unique
class ModuleType(Enum):
    NRPS = 1
    PKS = 2

    @staticmethod
    def from_string(label: str) -> "ModuleType":
        for value in ModuleType:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown module type: {label}")


@unique
class PKSModuleSubtype(Enum):
    PKS_CIS = 1
    PKS_TRANS = 2
    PKS_ITER = 3

    @staticmethod
    def from_string(label: str) -> "PKSModuleSubtype":
        for value in PKSModuleSubtype:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown PKS module subtype: {label}")


@unique
class NRPSDomainType(Enum):
    A = 1
    C = 2
    PCP = 3
    E = 4
    nMT = 5
    CYC = 6
    TE = 7
    TD = 8
    UNKNOWN = 9
    OX = 10
    CAL = 11

    @staticmethod
    def from_string(label: str) -> "NRPSDomainType":
        for value in NRPSDomainType:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown NRPS domain type: {label}")


@unique
class PKSDomainType(Enum):
    AT = 1
    KS = 2
    ACP = 3
    KR = 4
    DH = 5
    ER = 6
    TE = 7
    TD = 8
    UNKNOWN = 9
    DUMMY_ALMT = 10  # Alpha-L-Methyltransferase
    DUMMY_AMT = 11  # Alpha-Methyltransferase
    DUMMY_SC = 12  # Smalles cyclase for creating pyran/furan rings
    DUMMY_ZDH = 13  # E-configured double bonds
    DUMMY_EDH = 14  # Z-configured double bonds
    DUMMY_AH = 15  # Alpha-hydroxylase
    DUMMY_GDH = 16  # Gamma-beta-dehydrogenase
    DUMMY_ZGDH = 17  # Z-Gamma-beta-dehydrogenase
    DUMMY_EGDH = 18  # E-Gamma-beta-dehydrogenase
    DUMMY_OMT = 19  # Beta-Hydroxymethyltransferase
    DUMMY_BMT = 20  # Beta-Methyltransferase
    CAL = 21
    DUMMY_AT = 22

    @staticmethod
    def from_string(label: str) -> "PKSDomainType":
        for value in PKSDomainType:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown PKS domain type: {label}")


class _Module:

    def __init__(
        self,
        nr: int,
        module_type: str,
        domains: List[Domain],
        *,
        module_subtype: Union[str, None] = None,
        starter: bool = False,
        terminator: bool = False,
    ) -> None:
        self.id = nr
        self.type = ModuleType.from_string(module_type)
        self.subtype = None

        if module_subtype is not None:
            if self.type.name == "PKS":
                self.subtype = PKSModuleSubtype.from_string(module_subtype)
            else:
                raise ValueError(
                    f"Module subtypes not supported for {self.type.name} modules."
                )

        self.domains = domains

        self.recognition_domain = None
        self.synthesis_domain = None
        self.carrier_domain = None
        self.termination_domain = None

        self.tailoring_domains = []

        self.is_broken = False
        self.is_starter_module = starter
        self.is_termination_module = terminator

        for domain in self.domains:
            if self.type.name == "NRPS":
                if domain.type.name not in [v.name for v in NRPSDomainType]:
                    raise ValueError(f"Unknown NRPS domain type {domain.type.name}")
            if self.type.name == "PKS":
                if domain.type.name not in [v.name for v in PKSDomainType]:
                    raise ValueError(f"Unknown PKS domain type {domain.type.name}")

            if domain.used and domain.active:
                if isinstance(domain, TailoringDomain):
                    if domain.type.name not in [
                        d.type.name for d in self.tailoring_domains
                    ]:
                        self.tailoring_domains.append(domain)
                    else:
                        raise ValueError(
                            f"Cannot have two used tailoring domains of type {domain.type.name} in one \
module. Remove domain or set the 'used' or 'active' flag to False"
                        )

                if isinstance(domain, RecognitionDomain):
                    if not self.recognition_domain:
                        self.recognition_domain = domain
                    else:
                        raise ValueError(
                            f"Cannot have more than one used recognition domain in one \
module. Remove a domain or set the 'used' or 'active' flag to False"
                        )

                if isinstance(domain, SynthesisDomain) and domain.is_elongating:
                    if not self.synthesis_domain:
                        self.synthesis_domain = domain
                    else:
                        raise ValueError(
                            "Cannot have more than one used and elongating synthesis domains \
in one module. Remove a domain, set the 'used' or 'active' flag to False, or set the 'is_elongating' flag to False."
                        )

                if isinstance(domain, CarrierDomain):
                    if not self.carrier_domain:
                        self.carrier_domain = domain
                    else:
                        raise ValueError(
                            "Cannot have more than one used carrier domain in one \
module. Remove a domain or set the 'used' or 'active' flag to False"
                        )

                if isinstance(domain, TerminationDomain):
                    if not self.termination_domain:
                        self.termination_domain = domain
                        if not self.is_termination_module:
                            self.termination_domain.used = False
                    else:
                        raise ValueError(
                            "Cannot have more than one used termination domain in one \
module. Remove a domain or set the 'used' or 'active' flag to False"
                        )
        # I would rather implement a "dont do anything"- broken module
        if not self.is_starter_module and not self.synthesis_domain:
            if "CYC" in [domain.type.name for domain in self.tailoring_domains]:
                self.synthesis_domain = SynthesisDomain("DUMMY_C")
            else:
                self.is_broken = True

        if (
            not self.is_starter_module
            and self.recognition_domain
            and self.recognition_domain.domain_name == "CAL"
        ):
            self.is_broken = True

        if module_subtype != "PKS_TRANS" and not self.recognition_domain:
            self.is_broken = True

        if self.is_termination_module and not self.termination_domain:
            self.termination_domain = TerminationDomain("DUMMY_TE")

        if not self.carrier_domain:
            self.is_broken = True

        if self.is_broken:
            for domain in self.domains:
                domain.used = False

        for domain in self.domains[:]:
            if "DUMMY" in domain.type.name:
                self.domains.remove(domain)

    def run_module(self, structure: Union[Structure, None] = None):
        raise NotImplementedError

    def get_tailoring_domain(self, domain_name: str) -> Union[TailoringDomain, None]:
        for domain in self.tailoring_domains:
            if domain.type.name == domain_name:
                return domain

        return

    def add_module_label(self, structure):
        for atom in structure.graph:
            if type(atom) == Atom and not atom.annotations.has_annotation("module_nr"):
                atom.annotations.add_annotation("module_nr", f"module_{self.id + 1:02}")

            if type(atom) == ClassCarrierDomain:
                for neighbour in structure.graph[atom]:
                    if neighbour.type == "S":
                        if not neighbour.annotations.has_annotation("module_nr"):
                            neighbour.annotations.add_annotation(
                                "module_nr", f"module_{self.id + 1:02}"
                            )
                        else:
                            neighbour.annotations.module_nr = f"module_{self.id + 1:02}"

    def do_pks_tailoring(self, structure: Structure) -> Structure:
        kr_domain = self.get_tailoring_domain("KR")
        dh_domain = self.get_tailoring_domain("DH")
        er_domain = self.get_tailoring_domain("ER")

        if kr_domain and kr_domain.active and kr_domain.used:
            assert kr_domain.subtype is not None
            structure, kr_tailored = kr_domain.do_tailoring(structure)
            if kr_tailored:
                if (
                    not kr_domain.subtype.name == "C1"
                    and not kr_domain.subtype.name == "C2"
                ):
                    if dh_domain and dh_domain.active and dh_domain.used:
                        structure, dh_tailored = dh_domain.do_tailoring(structure)
                        if dh_tailored:
                            if er_domain and er_domain.active and er_domain.used:
                                structure, er_tailored = er_domain.do_tailoring(
                                    structure
                                )
                                if not er_tailored:
                                    er_domain.used = False

                        else:
                            dh_domain.used = False
                            if er_domain:
                                er_domain.used = False
            else:
                kr_domain.used = False
                if dh_domain:
                    dh_domain.used = False
                if er_domain:
                    er_domain.used = False
        else:
            if dh_domain:
                dh_domain.used = False
            if er_domain:
                er_domain.used = False

        return structure

    def do_nrps_tailoring(self, structure: Structure) -> Structure:
        e_domain = self.get_tailoring_domain("E")
        n_mt_domain = self.get_tailoring_domain("nMT")
        cyc_domain = self.get_tailoring_domain("CYC")
        ox_domain = self.get_tailoring_domain("OX")

        if e_domain and e_domain.active and e_domain.used:
            structure, epimerized = e_domain.do_tailoring(structure)
            if not epimerized:
                e_domain.used = False
        if n_mt_domain and n_mt_domain.active and n_mt_domain.used:

            structure, methylated = n_mt_domain.do_tailoring(structure)
            if not methylated:
                n_mt_domain.used = False

        if cyc_domain and cyc_domain.active and cyc_domain.used:
            structure, cyclised = cyc_domain.do_tailoring(structure)
            if not cyclised:
                cyc_domain.used = False

            if ox_domain and ox_domain.active and ox_domain.used:
                structure, oxidated = ox_domain.do_tailoring(structure)
                if not oxidated:
                    ox_domain.used = False

        elif ox_domain:
            ox_domain.used = False

        return structure

    def release_chain(self, structure: Structure) -> Structure:
        assert structure
        assert self.termination_domain

        released_structure = self.termination_domain.release_chain(structure)

        return released_structure


class LinearPKSModule(_Module):

    def __init__(self, nr, domains, starter=False, terminator=False):
        super().__init__(
            nr,
            "PKS",
            domains,
            starter=starter,
            terminator=terminator,
            module_subtype="PKS_CIS",
        )

    def run_module(self, structure: Union[Structure, None] = None) -> Structure:
        if self.is_broken:
            return structure
        else:
            if self.recognition_domain:
                if structure is None:
                    assert self.is_starter_module

                    if self.recognition_domain.domain_name != 'CAL':
                        if not self.recognition_domain.substrate.starter_monomer:
                            raise ValueError("Substrate of cis-AT PKS starter module is not a starter substrate")
                        structure = self.recognition_domain.substrate.starter_monomer.attach_to_acp()

                    else:
                        starter_unit = read_smiles(
                            self.recognition_domain.substrate.smiles
                        )
                        label_nrp_central_chain(starter_unit)
                        structure = attach_to_domain_nrp(starter_unit)

                else:
                    structure = self.synthesis_domain.do_elongation(
                        structure, self.recognition_domain.substrate
                    )

                structure = self.do_pks_tailoring(structure)
            self.add_module_label(structure)
            return structure


class IterativePKSModule(_Module):
    def __init__(
        self, nr, domains, starter=False, terminator=False, iterations: int = 1
    ) -> None:
        super().__init__(
            nr,
            "PKS",
            domains,
            starter=starter,
            terminator=terminator,
            module_subtype="PKS_ITER",
        )
        self.iterations = iterations

    def run_module(self, structure=None):
        if self.is_broken:
            return structure
        else:
            for i in range(0, self.iterations):
                if self.recognition_domain:
                    # make new substrate to enable reuse of substrate
                    substrate = PKSSubstrate(self.recognition_domain.substrate.name)
                    if structure is None:
                        assert self.is_starter_module
                        if not self.recognition_domain.substrate.starter_monomer:
                            substrate = PKSSubstrate("ACETYL_COA")
                        structure = substrate.starter_monomer.attach_to_acp()

                    else:
                        structure = self.synthesis_domain.do_elongation(
                            structure, substrate
                        )

                    structure = self.do_pks_tailoring(structure)
            self.add_module_label(structure)
            return structure


class TransATPKSModule(_Module):
    def __init__(
        self, nr, domains, substrate_name="MALONYL_COA", starter=False, terminator=False
    ) -> None:
        super().__init__(
            nr,
            "PKS",
            domains,
            starter=starter,
            terminator=terminator,
            module_subtype="PKS_TRANS",
        )
        if not self.recognition_domain:
            self.recognition_domain = RecognitionDomain("DUMMY_AT", substrate_name)

    def do_pks_tailoring(self, structure: Structure) -> Structure:

        kr_domain = self.get_tailoring_domain("KR")
        if not kr_domain:
            kr_domain = self.get_tailoring_domain("DUMMY_KR")
        dh_domain = self.get_tailoring_domain("DH")
        if not dh_domain:
            dh_domain = self.get_tailoring_domain("DUMMY_DH")
        edh_domain = self.get_tailoring_domain("EDH")
        if not edh_domain:
            edh_domain = self.get_tailoring_domain("DUMMY_EDH")
        zdh_domain = self.get_tailoring_domain("ZDH")
        if not zdh_domain:
            zdh_domain = self.get_tailoring_domain("DUMMY_ZDH")
        er_domain = self.get_tailoring_domain("ER")
        if not er_domain:
            er_domain = self.get_tailoring_domain("DUMMY_ER")
        almt_domain = self.get_tailoring_domain("ALMT")
        if not almt_domain:
            almt_domain = self.get_tailoring_domain("DUMMY_ALMT")
        amt_domain = self.get_tailoring_domain("AMT")
        if not amt_domain:
            amt_domain = self.get_tailoring_domain("DUMMY_AMT")
        emo_domain = self.get_tailoring_domain("EMO")
        if not amt_domain:
            emo_domain = self.get_tailoring_domain("DUMMY_EMO")
        sc_domain = self.get_tailoring_domain("SC")
        if not sc_domain:
            sc_domain = self.get_tailoring_domain("DUMMY_SC")
        ah_domain = self.get_tailoring_domain("AH")
        if not ah_domain:
            ah_domain = self.get_tailoring_domain("DUMMY_AH")
        gdh_domain = self.get_tailoring_domain("GDH")
        if not gdh_domain:
            gdh_domain = self.get_tailoring_domain("DUMMY_GDH")
        zgdh_domain = self.get_tailoring_domain("ZGDH")
        if not zgdh_domain:
            zgdh_domain = self.get_tailoring_domain("DUMMY_ZGDH")
        egdh_domain = self.get_tailoring_domain("EGDH")
        if not egdh_domain:
            egdh_domain = self.get_tailoring_domain("DUMMY_EGDH")
        omt_domain = self.get_tailoring_domain("OMT")
        if not omt_domain:
            omt_domain = self.get_tailoring_domain("DUMMY_OMT")
        bmt_domain = self.get_tailoring_domain("BMT")
        if not bmt_domain:
            bmt_domain = self.get_tailoring_domain("DUMMY_BMT")

        if ah_domain and ah_domain.active and ah_domain.used:
            structure, ah_tailored = ah_domain.do_tailoring(structure)
            if not ah_tailored:
                ah_domain.used = False
        if amt_domain and amt_domain.active and amt_domain.used:
            structure, amt_tailored = amt_domain.do_tailoring(structure)
            if not amt_tailored:
                amt_domain.used = False
            if emo_domain and emo_domain.active and emo_domain.used:
                structure, emo_tailored = emo_domain.do_tailoring(structure)
                if not emo_tailored:
                    emo_domain.used = False
            structure.refresh_structure()
        if almt_domain and almt_domain.active and almt_domain.used:
            structure, almt_tailored = almt_domain.do_tailoring(structure)
            if not almt_tailored:
                almt_domain.used = False
            if emo_domain and emo_domain.active and emo_domain.used:
                structure, emo_tailored = emo_domain.do_tailoring(structure)
                if not emo_tailored:
                    emo_domain.used = False
        if kr_domain and kr_domain.active and kr_domain.used:
            assert kr_domain.subtype is not None

            structure, kr_tailored = kr_domain.do_tailoring(structure)
            if (
                not kr_tailored
                or kr_domain.subtype.name == "C1"
                or kr_domain.subtype.name == "C2"
            ):
                kr_domain.used = False
                if omt_domain:
                    omt_domain.used = False
                if dh_domain:
                    dh_domain.used = False
                if bmt_domain:
                    bmt_domain.used = False
                if er_domain:
                    er_domain.used = False
            if (
                not kr_domain.subtype.name == "C1"
                and not kr_domain.subtype.name == "C2"
            ):
                if bmt_domain and bmt_domain.active and bmt_domain.used:
                    structure, bmt_tailored = bmt_domain.do_tailoring(structure)
                    if not bmt_tailored:
                        bmt_domain.used = False
                if omt_domain and omt_domain.active and omt_domain.used:
                    structure, omt_tailored = omt_domain.do_tailoring(structure)
                    if not omt_tailored:
                        omt_domain.used = False
                elif dh_domain and dh_domain.active and dh_domain.used:
                    structure, dh_tailored = dh_domain.do_tailoring(structure)
                    if not dh_tailored:
                        dh_domain.used = False
                        if er_domain:
                            er_domain.used = False
                    else:
                        if er_domain and er_domain.active and er_domain.used:
                            structure, er_tailored = er_domain.do_tailoring(structure)
                            if not er_tailored:
                                er_domain.used = False
                elif edh_domain and edh_domain.active and edh_domain.used:
                    structure, edh_tailored = edh_domain.do_tailoring(structure)
                    if not edh_tailored:
                        edh_domain.used = False

                        if er_domain:
                            er_domain.used = False

                    else:
                        if er_domain and er_domain.active and er_domain.used:
                            structure, er_tailored = er_domain.do_tailoring(structure)
                            if not er_tailored:
                                er_domain.used = False
                elif zdh_domain and zdh_domain.active and zdh_domain.used:
                    structure, zdh_tailored = zdh_domain.do_tailoring(structure)
                    if not zdh_tailored:
                        zdh_domain.used = False
                        if er_domain:
                            er_domain.used = False
                    else:
                        if er_domain and er_domain.active and er_domain.used:
                            structure, er_tailored = er_domain.do_tailoring(structure)
                            if not er_tailored:
                                er_domain.used = False
                elif gdh_domain and gdh_domain.active and gdh_domain.used:
                    structure, gdh_domain_tailored = gdh_domain.do_tailoring(structure)
                    if not gdh_domain_tailored:
                        gdh_domain.used = False
                        if er_domain:
                            er_domain.used = False
                    else:
                        if er_domain and er_domain.active and er_domain.used:
                            structure, er_tailored = er_domain.do_tailoring(structure)
                            if not er_tailored:
                                er_domain.used = False
                elif egdh_domain and egdh_domain.active and egdh_domain.used:
                    structure, egdh_domain_tailored = egdh_domain.do_tailoring(
                        structure
                    )
                    if not egdh_domain_tailored:
                        egdh_domain.used = False
                        if er_domain:
                            er_domain.used = False
                    else:
                        if er_domain and er_domain.active and er_domain.used:
                            structure, er_tailored = er_domain.do_tailoring(structure)
                            if not er_tailored:
                                er_domain.used = False
                elif zgdh_domain and zgdh_domain.active and zgdh_domain.used:
                    structure, zgdh_domain_tailored = zgdh_domain.do_tailoring(
                        structure
                    )
                    if not zgdh_domain_tailored:
                        zgdh_domain.used = False
                        if er_domain:
                            er_domain.used = False
                    else:
                        if er_domain and er_domain.active and er_domain.used:
                            structure, er_tailored = er_domain.do_tailoring(structure)
                            if not er_tailored:
                                er_domain.used = False
        if sc_domain and sc_domain.active and sc_domain.used:
            structure, sc_tailored = sc_domain.do_tailoring(structure)
            if not sc_tailored:
                sc_domain.used = False

        structure.refresh_structure()
        return structure

    def run_module(self, structure: Union[Structure, None] = None) -> Structure:
        if self.is_broken:
            return structure
        else:
            if self.recognition_domain:
                if structure is None:
                    assert self.is_starter_module
                    structure = (
                        self.recognition_domain.substrate.starter_monomer.attach_to_acp()
                    )
                else:
                    structure = self.synthesis_domain.do_elongation(
                        structure, self.recognition_domain.substrate
                    )
                structure = self.do_pks_tailoring(structure)
            self.add_module_label(structure)
            return structure


class NRPSModule(_Module):

    def __init__(self, nr, domains, starter=False, terminator=False):
        super().__init__(nr, "NRPS", domains, starter=starter, terminator=terminator)

    def run_module(self, structure=None) -> Structure:
        if self.is_broken:
            return structure
        else:
            if self.recognition_domain:
                if structure is None:
                    assert self.is_starter_module
                    starter_unit = read_smiles(self.recognition_domain.substrate.smiles)
                    label_nrp_central_chain(starter_unit)
                    structure = attach_to_domain_nrp(starter_unit)
                else:
                    structure = self.synthesis_domain.do_elongation(
                        structure, self.recognition_domain.substrate
                    )

                structure = self.do_nrps_tailoring(structure)
            self.add_module_label(structure)
            return structure



import math
import pygame

pygame.font.init()

BACKGROUND_COLOUR = (255, 255, 255)
BUTTON_PANEL_COLOUR = (255, 255, 255)

BLACK = (0, 0, 0)
WHITE = (255, 255, 255)
RED = (198, 0, 0)

SIZE = 1300, 800
WIDTH, HEIGHT = SIZE

PANEL_LOCATION = (0, HEIGHT / 2)
PANEL_HEIGHT = HEIGHT / 2
PANEL_SIZE = (WIDTH, PANEL_HEIGHT)

KR_BUTTON_SIZE = 100

BUTTON_PANEL = pygame.Rect(PANEL_LOCATION, PANEL_SIZE)
# FONT = "couriernew"
FONT = "helvetica"
SUBSTRATE_FONT = "arial"

SQUARE_WIDTH = int(HEIGHT / 13)

PIECE_PADDING = int(math.ceil(3 * SQUARE_WIDTH / 100))
SYMBOL_PADDING = int(math.ceil(28 * SQUARE_WIDTH / 100))

MODULE_SPACING = 40
MODULE_PADDING = 20

GENE_PADDING = 50
GENE_LABEL_SIZE = 60
GENE_SPACING = 50
GENE_HEIGHT = 20
DOMAIN_SIZE = 30

BUTTON_COLOUR = (200, 200, 200)
BUTTON_HIGHLIGHT_COLOUR = (230, 230, 230)

GENE_COLOUR = (200, 200, 200)
GENE_HIGHLIGHT_COLOUR = (220, 220, 220)
MODULE_COLOUR = (240, 240, 240)
MODULE_HIGHLIGHT_COLOUR = (245, 245, 245)
MODULE_LINE_COLOUR = (100, 100, 100)


DOMAIN_BUTTON_SIZE = 40

BUTTON_FONT = pygame.font.SysFont(FONT, 20, bold=True)
BUTTON_TEXT_COLOUR = BLACK
SUBSTRATE_BUTTON_SIZE = 140
SUBSTRATE_BUTTONS_PER_LINE = 6
SUBSTRATE_BUTTON_PADDING = 30
SUBSTRATE_GROUP_BUTTON_SIZE = (int(WIDTH * 0.05), int(HEIGHT / 25))
SUBSTRATE_GROUP_BUTTON_PADDING = int(WIDTH * 0.02)
SUBSTRATE_GROUP_BUTTONS_PER_LINE = 6

RENDER_WINDOW_SIZE = (1000, 800)

FATTY_ACID_IMAGE_SIZE = (200, 200)

PIECE_BUTTON_WIDTH = int(HEIGHT / 6)

PIECE_BUTTON_HEIGHT = int(HEIGHT / 25)

CONDITION_SIZE = int(math.ceil(SQUARE_WIDTH / 5))

REPLACE_TEXT_FONT = pygame.font.SysFont(FONT, 14, bold=False)

REPLACE_TEXT_1 = REPLACE_TEXT_FONT.render("This file destination already exists.", True, BLACK)
REPLACE_TEXT_2 = REPLACE_TEXT_FONT.render("Are you sure you want to overwrite?", True, BLACK)

FAILED_TEXT_1 = REPLACE_TEXT_FONT.render("Failed to render cluster.", True, BLACK)
FAILED_TEXT_2 = REPLACE_TEXT_FONT.render("Check if starter units were used as elongation units.", True, BLACK)
FAILED_TEXT_3 = REPLACE_TEXT_FONT.render("Failed to render products.", True, BLACK)




from raichu.drawing.colours import (
    OUTLINE_COLOURS,
    FILL_COLOURS,
    DOMAIN_ABBREVIATIONS,
    TRANSPARENT_OUTLINE_COLOURS,
    TRANSPARENT_FILL_COLOURS,
    TRANSPARENT_TEXT_COLOURS,
    TEXT_COLOUR,
)


def make_circle(x_coord, y_coord, domain):
    """Easy function to draw circle for the domain visualization. Returns
    matplotlib.patches.circle object of a circle with radius 4

    x_coord: int, x-coordinate of the center of the circle to be drawn
    domain_type: str, PKS domain type (ACP, KS, AT, KR, DH or ER)
    """
    if domain.used:
        colour = FILL_COLOURS[domain.type.name]
        outline_colour = OUTLINE_COLOURS[domain.type.name]
    else:
        colour = TRANSPARENT_FILL_COLOURS[domain.type.name]
        outline_colour = TRANSPARENT_OUTLINE_COLOURS[domain.type.name]

    circle = f"""<circle r="14" cx="{int(x_coord)}" cy="{int(y_coord)}" stroke="{outline_colour}"
       fill="{colour}" stroke-width="1"/>"""
    return circle


def draw_rectangle(gene_name, x, y, width, height=12, text_colour="white"):
    rectangle = (
        f"""<rect x="{x}" y="{y}" width="{width}" height="{height}" fill="black" />"""
    )
    text_x = x + width / 2
    text_y = y + height / 2
    text = f"""<text x="{text_x}" y="{text_y}" fill="{text_colour}" text-anchor="middle" font-family="verdana" font-size = "{8}">\
<tspan y="{text_y}" dy="0.35em">{gene_name}</tspan></text>"""

    svg_rect = f"""<g id="gene_banner_{gene_name}">"""
    svg_rect += rectangle
    svg_rect += text
    svg_rect += "</g>"

    return svg_rect


def draw_line(start_x, end_x, y=4):
    line = f"""<line x1="{start_x}" x2="{end_x}" y1="{y}" y2="{y}" stroke-width="2" stroke="black" />"""
    return line


def make_module_text(start_x, end_x, y, module_nr, offset_1=True):
    if offset_1:
        module_nr += 1
    text_x = start_x + (end_x - start_x) / 2
    text = f"""<text x="{text_x}" y="{y}" fill="'black" text-anchor="middle" font-family="verdana" font-size = "{12}">\
<tspan y="{y}" dy="0.35em">Module {module_nr}</tspan></text>"""
    return text


def draw_bubbles(cluster, widths, delta_x=29, bubble_height=80, min_gene_padding=20, min_module_padding=10):
    x = 30.0
    x_bubbles = 30.0

    cp_positions = []
    cp_positions_bubbles = []
    previous_space_right = 0.0
    previous_cp_position = 0.0

    for i, module in enumerate(cluster.modules):
        current_y = bubble_height
        for j, domain in enumerate(module.domains):

            # Take note of whether the domain is part of a new gene (modules can be split across multiple genes)
            new_gene = False
            if j < len(module.domains) - 1:
                current_gene = domain.gene
                next_gene = module.domains[j + 1].gene
                if current_gene != next_gene:
                    new_gene = True

            # Move up bubbles representing unknown and tailoring domains

            if domain.supertype.name == "UNKNOWN" or domain.supertype.name == "TAILORING":

                if current_y == bubble_height - 7:
                    level_change = False
                else:
                    level_change = True

                current_y = bubble_height - 7

            else:
                if current_y == bubble_height:
                    level_change = False
                else:
                    level_change = True
                current_y = bubble_height

            # If there is a level change, you can place the bubble closer to the previous one

            if level_change and j != 0:
                x_bubbles -= 1
                x -= 1

            # Only note the position of the cp domain if the module is not broken

            if domain.supertype.name == "CARRIER" and domain.used and not module.is_broken:
                cp_positions_bubbles.append(x_bubbles)

                # Record how much space the structures take up left and right

                current_space_left, current_space_right = widths[i]

                minimum_x = previous_cp_position + current_space_left + previous_space_right

                # If structures take up more space than the bubbles, shift the CP domain to the right

                x = max([minimum_x, x])
                cp_positions.append(x)
                previous_space_right = current_space_right
                previous_cp_position = x

            x_bubbles += delta_x
            x += delta_x

            # If a new gene is introduced, move the bubbles over

            if new_gene:
                x_bubbles += min_gene_padding
                x += min_gene_padding

        x += min_module_padding
        x_bubbles += min_module_padding

    module_shifts = []

    # Calculate how much each module has to shift

    for i, cp_position in enumerate(cp_positions):
        module_shift = cp_positions[i] - cp_positions_bubbles[i]
        for shift in module_shifts:
            module_shift -= shift
        module_shifts.append(module_shift)

    if module_shifts:

        current_x = 30.0 + module_shifts[0]

    else:
        current_x = 30.0

    circles = []
    texts = []
    cp_positions = []
    lines = []
    current_gene = None
    gene_start = 0
    gene_end = 0
    gene_rects = []
    genes = []
    module_texts = []

    correct_modules = 0

    for i, module in enumerate(cluster.modules):

        start_x = current_x - 15
        current_y = bubble_height

        for j, domain in enumerate(module.domains):
            if domain.gene not in genes:
                genes.append(domain.gene)
            if current_gene and domain.gene != current_gene:
                width = gene_end - gene_start
                gene_rects.append(
                    draw_rectangle(current_gene, gene_start, bubble_height - 75, width)
                )
                gene_start = current_x - 15
            elif not current_gene:
                gene_start = current_x - 15

            current_gene = domain.gene

            new_gene = False
            if j < len(module.domains) - 1:
                next_gene = module.domains[j + 1].gene
                if current_gene != next_gene:
                    new_gene = True

            abbreviation = DOMAIN_ABBREVIATIONS.get(domain.type.name)
            if abbreviation is None:
                abbreviation = DOMAIN_ABBREVIATIONS.get(domain.domain_name)
                if abbreviation is None:
                    abbreviation = ""

            if (
                domain.supertype.name == "UNKNOWN"
                or domain.supertype.name == "TAILORING"
            ):
                if current_y == bubble_height - 7:
                    level_change = False
                else:
                    level_change = True

                current_y = bubble_height - 7
            else:
                if current_y == bubble_height:
                    level_change = False
                else:
                    level_change = True
                current_y = bubble_height

            if level_change and j != 0:
                current_x -= 1

            circles.append(make_circle(current_x, current_y, domain))

            text_colour = TEXT_COLOUR
            if not domain.used:
                text_colour = TRANSPARENT_TEXT_COLOURS[domain.type.name]

            text = f"""<text x="{current_x}" y="{current_y}" fill="{text_colour}" text-anchor="middle" font-family="verdana" font-size = "{13}">\
<tspan y="{current_y}" dy="0.35em">{abbreviation}</tspan></text>"""

            texts.append(text)
            if domain.supertype.name == "CARRIER" and domain.used:
                cp_positions.append((current_x, current_y))

            gene_end = current_x + 15

            current_x += delta_x
            if new_gene:

                current_x += min_gene_padding

        # Shift modules based on width of structures

        module_shift = 0.0

        if not cluster.modules[i].is_broken:
            correct_modules += 1

        if i != len(cluster.modules) - 1 and not cluster.modules[i + 1].is_broken:

            module_shift = module_shifts[correct_modules]

        end_x = current_x - 15

        if not cluster.modules[i].is_broken:

            lines.append(draw_line(start_x, end_x, y=bubble_height - 26))
            module_texts.append(make_module_text(start_x, end_x, bubble_height - 34, correct_modules - 1))

        current_x += min_module_padding + module_shift

    if gene_start and gene_end:
        width = gene_end - gene_start
        gene_rects.append(
            (draw_rectangle(current_gene, gene_start, bubble_height - 75, width))
        )

    svg = ""

    svg += f"""<g id="domain_circles">\n"""
    for i, circle in enumerate(circles):
        text = texts[i]
        svg += f"""<g id="domain_bubble_{i}">\n"""
        svg += f"{circle}\n"
        svg += f"{text}\n"
        svg += "</g>\n"
    svg += "</g>\n"

    svg += f"""<g id="gene_rectangles">\n"""
    for rect in gene_rects:
        svg += f"{rect}\n"
    svg += "</g>\n"

    svg += f"""<g id="module_labels">\n"""
    for i, line in enumerate(lines):
        svg += f"""<g id="module_label_{i}">"""
        module_text = module_texts[i]
        svg += f"{line}\n"
        svg += f"{module_text}\n"
        svg += "</g>\n"
    svg += "</g>\n"

    return svg, cp_positions, current_x



from pikachu.reactions.functional_groups import combine_structures
from raichu.data.molecular_moieties import THIOESTERBOND, THIOESTERBOND_OXYGEN_INSERTED
from raichu.reactions.general import label_rest_groups
from raichu.reactions.pks_sidechain_chirality import set_sidechain_chirality


def pks_elongation(chain_intermediate, elongation_monomer):
    """
    Returns a Structure object of the PK chain after a single elongation step
    elongation step.

    structure: PIKAChU Structure object of the RC(=O)S PK intermediate before
    the elongation step
    elongation_monomer: ['SMILES_elongation_unit', index_c_to_c, index_c_to_s']
    """

    # If this is the first elongation reaction on the starter unit, define
    # central chain atoms in the starter unit
    assert any(atom.annotations.in_central_chain for atom in chain_intermediate.graph)

    h_to_remove_1 = None
    h_to_remove_2 = None

    for atom in elongation_monomer.c_to_pk_intermediate.neighbours:
        if atom.type == 'H':
            h_to_remove_1 = atom
            break

    for atom in elongation_monomer.c_to_s.neighbours:
        if atom.type == 'H':
            h_to_remove_2 = atom
            break

    assert h_to_remove_1 and h_to_remove_2

    # If the elongation unit contains a wildcard, add a number to the
    # '*' atom to display in the structure drawing
    label_rest_groups(elongation_monomer.structure, chain_intermediate)

    thioester_bonds = find_bonds(chain_intermediate, THIOESTERBOND)
    if len(thioester_bonds) == 0:
        thioester_bonds = find_bonds(chain_intermediate, THIOESTERBOND_OXYGEN_INSERTED)

    assert len(thioester_bonds) == 1
    for bond in thioester_bonds:
        if bond.atom_1.type == 'S':
            s_pkchain = bond.atom_1
            c_pkchain = bond.atom_2
        elif bond.atom_2.type == 'S':
            s_pkchain = bond.atom_2
            c_pkchain = bond.atom_1

    new_structure = combine_structures([elongation_monomer.structure, chain_intermediate])

    # Remove the Hs in the malonyl derivative in order to add it to the pk chain
    elongation_monomer.structure.remove_atom(new_structure.get_atom(h_to_remove_1))
    elongation_monomer.structure.remove_atom(new_structure.get_atom(h_to_remove_2))

    # Find thioester bond in growing chain, define Cs to attach new unit

    # Breaking the thioester bond in the PK chain
    for atom in chain_intermediate.graph:
        atom.hybridise()
    for bond in thioester_bonds[:]:
        chain_intermediate.break_bond(bond)

    chain_intermediate.get_connectivities()
    chain_intermediate.set_connectivities()
    chain_intermediate.set_atom_neighbours()
    chain_intermediate.make_bond_lookup()

    # Refresh malonylunit

    elongation_monomer.structure.refresh_structure()

    # Combining structures PK chain and elongation unit into one Structure object
    pk_chain_and_malonyl = (elongation_monomer.structure, chain_intermediate)

    combined = combine_structures(pk_chain_and_malonyl)

    combined.get_connectivities()
    combined.set_connectivities()
    combined.set_atom_neighbours()
    combined.make_bond_lookup()

    # Adding the bonds to form the new molecule after the single elongation step
    new_bond_nr = combined.find_next_bond_nr()
    combined.make_bond(elongation_monomer.c_to_pk_intermediate, c_pkchain, new_bond_nr)
    new_bond_nr = combined.find_next_bond_nr()

    combined.make_bond(elongation_monomer.c_to_s, s_pkchain, new_bond_nr)
    combined.get_connectivities()
    combined.set_connectivities()
    combined.set_atom_neighbours()
    combined.make_bond_lookup()
    for bond_nr, bond in combined.bonds.items():
        bond.set_bond_summary()

    for atom in combined.graph:
        atom.annotations.c2_acid = False

    set_sidechain_chirality(combined)

    return combined


def find_bonds(structure, bond_type):
    """
    Returns a list of Pikachu.bond objects of the bonds of the indicated type
    in the structure of interest

    structure: Pikachu.structure object of the molecule of interest
    bond_type: BondDefiner object indicating the bond type that is searched
    in the indicated structure
    """
    for atom in structure.graph:
        atom.get_connectivity()
    locations = structure.find_substructures(bond_type.structure)
    bonds = []
    for match in locations:
        atom_1 = match.atoms[bond_type.atom_1]
        atom_2 = match.atoms[bond_type.atom_2]
        bond = structure.bond_lookup[atom_1][atom_2]
        bonds.append(bond)

    return bonds



import pygame
from interactive.style import BLACK, GENE_COLOUR, GENE_LABEL_SIZE, GENE_SPACING, GENE_PADDING, \
    GENE_HEIGHT, BACKGROUND_COLOUR, FONT, MODULE_PADDING, DOMAIN_SIZE, GENE_HIGHLIGHT_COLOUR
from interactive.module import Module

GENE_X_START = 30


class Gene:
    def __init__(self, screen, gene_number):
        self.name = None
        self.gene_number = gene_number
        self.screen = screen
        self.modules = []
        self.gene_length = 0
        self.x = GENE_PADDING
        self.y = GENE_PADDING + self.gene_number * GENE_SPACING
        self.width = 0
        self.height = GENE_HEIGHT
        self.dna_coords = None
        self.strand = None

        self.rectangle = None
        self.set_rectangle()

        self.font_size = self.height - 10

        self.text_x = self.x + 2
        self.text_y = self.y + int((self.height - self.font_size) / 2)

        self.font = pygame.font.SysFont(FONT, self.font_size, bold=False)

        self.selected = False

        self.insertion_point = None

    def sort_modules(self):
        self.modules.sort(key=lambda module: module.dna_coords[0])
        if self.strand == -1:
            self.modules.reverse()

        for i, module in enumerate(self.modules):
            module.id = i

    def set_rectangle(self):
        self.y = GENE_PADDING + self.gene_number * GENE_SPACING
        self.width = GENE_LABEL_SIZE
        for module in self.modules:
            self.width += module.width
            self.width += MODULE_PADDING

        self.width += MODULE_PADDING

        self.rectangle = pygame.Rect(self.x, self.y, self.width, self.height)

    def draw(self, mouse):

        for module in self.modules:
            module.set_rectangle()

        self.set_rectangle()

        self.text_x = self.x + 2
        self.text_y = self.y + int((self.height - self.font_size) / 2)

        colour = GENE_COLOUR
        if self.rectangle.collidepoint(mouse) or self.selected:
            for module in self.modules:
                if module.rectangle.collidepoint(mouse) or module.selected:
                    break
            colour = GENE_HIGHLIGHT_COLOUR

        pygame.draw.rect(self.screen, colour, self.rectangle)
        pygame.draw.rect(self.screen, BLACK, self.rectangle, 1)

        rendered_text = self.font.render(self.name, True, BLACK)
        self.screen.blit(rendered_text, (self.text_x, self.text_y))

        for module in self.modules:
            module.draw(mouse)

        if self.insertion_point:
            self.insertion_point.draw(self.screen)

    def erase(self):
        height_difference = DOMAIN_SIZE - self.height
        rectangle = pygame.Rect(self.x, self.y - height_difference / 2 - 20, self.width, self.height + height_difference + 20)
        pygame.draw.rect(self.screen, BACKGROUND_COLOUR, rectangle)
        #pygame.draw.rect(self.screen, BACKGROUND_COLOUR, rectangle, 5)

    def adjust_module_indices(self, insertion_point):
        for module in self.modules:
            if module.id >= insertion_point:
                module.id += 1
                module.set_rectangle()

    def remove_module(self, module_nr):
        self.modules.pop(module_nr)

        for module in self.modules:
            if module.id > module_nr:
                module.id -= 1

    def get_domains(self):
        domains = []
        for module in self.modules:
            for domain in module.domains:
                domains.append(domain)

        return domains

    def add_module(self, insertion_point, module_type):

        self.adjust_module_indices(insertion_point)
        new_module = Module(self.screen, insertion_point, module_type, self)

        starter_module = False
        if insertion_point == 0 and self.gene_number == 0:
            starter_module = True

        domains = None

        if module_type == 'NRPS' and starter_module:
            domains = ['A', 'PCP']
        elif module_type == 'NRPS' and not starter_module:
            domains = ['C', 'A', 'PCP']
        elif module_type == 'PKS' and starter_module:
            domains = ['AT', 'ACP']
        elif module_type == 'PKS' and not starter_module:
            domains = ['KS', 'AT', 'ACP']

        assert domains

        for domain in domains:
            new_module.add_domain(domain)

        self.modules.insert(insertion_point, new_module)




